import argparse
import calendar
import logging
import os
import time
from datetime import datetime

from spade.message import Message
from spade.template import Template

from smia.logic.exceptions import CriticalError
from smia.utilities import properties_file_utils
from smia.utilities.smia_general_info import SMIAGeneralInfo

class GeneralUtils:
    """
    This class contains some general utils to ben used by any module.
    """


    @staticmethod
    def configure_logging():
        """
        This method configures the logging to be used by all modules. It specifies different colors to improve the
        readability of the console and adds new levels to the printouts related to ACL and interaction messages.
        """
        asset_level_num = 35
        fipa_acl_level_num = 36

        if ((logging.getLevelName("ASSETINFO")) == asset_level_num or
                (logging.getLevelName("ACLINFO") == fipa_acl_level_num)):
            # In this case the logging is already configured
            return

        logging.addLevelName(asset_level_num, "ASSETINFO")
        logging.addLevelName(fipa_acl_level_num, "ACLINFO")

        def assetinfo(self, message, *args, **kwargs):
            if self.isEnabledFor(asset_level_num):
                self._log(asset_level_num, message, args, **kwargs)

        logging.Logger.assetinfo = assetinfo

        def aclinfo(self, message, *args, **kwargs):
            if self.isEnabledFor(fipa_acl_level_num):
                self._log(fipa_acl_level_num, message, args, **kwargs)

        logging.Logger.aclinfo = aclinfo

        handler = logging.StreamHandler()
        formatter = GeneralUtils.ColoredFormatter('%(levelname)s: %(message)s')
        handler.setFormatter(formatter)
        logging.getLogger('').addHandler(handler)
        logging.getLogger('').setLevel(logging.INFO)  # Set the default logging level

        # Create a file handler
        file_handler = logging.FileHandler(SMIAGeneralInfo.LOG_FOLDER_PATH + '/' + SMIAGeneralInfo.SMIA_LOG_FILENAME)
        file_handler.setFormatter(logging.Formatter(GeneralUtils.ColoredFormatter.FORMAT_COMPLEX))
        logging.getLogger('').addHandler(file_handler)

    class ColoredFormatter(logging.Formatter):
        """
        This class contains the format of all the levels of the logging, including the color of each of them.
        """
        FORMAT_SIMPLE = "%(asctime)s [%(name)s] [%(levelname)s] %(message)s"
        FORMAT_COMPLEX = "%(asctime)s [%(name)s] [%(levelname)s] %(message)s line:%(lineno)d"
        RESET = '\x1b[0m'

        COLORS = {
            logging.DEBUG: '\x1b[94m' + FORMAT_SIMPLE + RESET,  # Blue
            logging.INFO: '\x1b[39;20m' + FORMAT_SIMPLE + RESET,  # White
            logging.WARNING: '\x1b[93m' + FORMAT_COMPLEX + RESET,  # Yellow
            logging.ERROR: '\x1b[91m' + FORMAT_COMPLEX + RESET,  # Red
            logging.CRITICAL: '\x1b[41m' + FORMAT_COMPLEX + RESET,  # White on Red
            35: '\x1b[38;2;255;150;20m' + FORMAT_SIMPLE + RESET,  # Purple (for the interaction level)
            36: '\x1b[38;2;0;255;255m' + FORMAT_SIMPLE + RESET  # Cyan (for the FIP-ACL level)
        }

        def format(self, record):
            log_fmt = self.COLORS.get(record.levelno)
            formatter = logging.Formatter(log_fmt)
            return formatter.format(record)

    @staticmethod
    def print_smia_banner():
        """
        This method prints the SMIA banner as a string. The banner has been created with Python 'art' library.
        """
        # The banner for the SMIA is set as string, avoiding installing 'art' library (which has been used to create it)
        banner_str = ("----------------------------------------------\n" +
                      "  ______    ____    ____   _____        _\n" +
                      ".' ____ \  |_   \  /   _| |_   _|      / \      \n" +
                      "| (___ \_|   |   \/   |     | |       / _ \     \n" +
                      " _.____`.    | |\  /| |     | |      / ___ \    \n" +
                      "| \____) |  _| |_\/_| |_   _| |_   _/ /   \ \_  \n" +
                      " \______.' |_____||_____| |_____| |____| |____| \n" +
                      # "                                                \n" +
                      "----------------------------------------------\n" +
                      "                                       v0.2.1 \n" +
                      "----------------------------------------------\n")
        print(banner_str)

    @staticmethod
    def update_aas_model(new_aas_model_file):
        """
        This method updates the AAS model file defined for the SMIA. It updates the global variable for the AAS model
        file name and the value within the properties file of the SMIA Archive.

        Args:
            new_aas_model_file (str): name of the file of the new AAS model to update SMIA.
        """
        # The global variable is updated
        SMIAGeneralInfo.CM_AAS_MODEL_FILENAME = new_aas_model_file

        # The properties file is updated
        file_name, file_ext = os.path.splitext(new_aas_model_file)
        properties_file_utils.set_aas_general_property('model.file', new_aas_model_file)
        properties_file_utils.set_aas_general_property('model.serialization', file_ext)

    @staticmethod
    def create_acl_template(performative, ontology):
        """
        This method creates a template aligned with FIPA-ACL standard.

        Args:
            performative(str): The performative of the template.
            ontology(str): The ontology of the template.

        Returns:
            spade.template.Template: a SPADE template object.
        """
        custom_template = Template()
        custom_template.set_metadata('performative', performative)
        custom_template.set_metadata('ontology', ontology)
        return custom_template

    @staticmethod
    def create_acl_msg(receiver, thread, performative, ontology, body):
        """
        This method creates an FIPA-ACL message.

        Args:
            receiver (str): The SPADE agent receiver of the ACL message.
            thread (str): The thread of the ACL message.:
            performative (str): The performative of the ACL message.
            ontology (str): The ontology of the ACL message.
            body: The body of the ACL message.

        Returns:
            spade.message.Message: SPADE message object FIPA-ACL-compliant.
        """
        msg = Message(to=receiver, thread=thread)
        msg.set_metadata('performative', performative)
        msg.set_metadata('ontology', ontology)

        msg.body = body  # TODO Pensar si iria tambien en metadatos o todo en el body
        return msg

    @staticmethod
    def get_sender_from_acl_msg(acl_msg):
        """
        This method returns the identifier of an agent from an ACL message, considering the suffixes that can be added
        by the XMPP server.

        Args:
            acl_msg (spade.message.Message): ACL message object.

        Returns:
            str: identifier of the sender of the message.
        """
        if '/' in str(acl_msg.sender):  # XMPP server can add a random string to differentiate the agent JID
            return str(acl_msg.sender).split('/')[0]
        else:
            return str(acl_msg.sender)

    @staticmethod
    def get_current_timestamp():
        """
        This method returns the current timestamp of the SMIA.

        Returns:
            int: current timestamp in milliseconds
        """
        return calendar.timegm(time.gmtime())

    @staticmethod
    def get_current_date_time():
        """
        This method returns the current DateTime of the SMIA. The DateTime is obtained from the UNIX timestamp.

        Returns:
            str: current DateTime.
        """
        return datetime.fromtimestamp(GeneralUtils.get_current_timestamp())


class CLIUtils:
    """This class contains utility methods related to the Command Line Interface."""

    @staticmethod
    def get_information_from_cli(cli_args):
        """
        This method gets all the information from the Command Line Interface (CLI). This information includes the
        initial configuration properties or the AAS model or. None of them is mandatory, but they must be consistent
        (i.e., if only the AAS model is offered, it must be an AASX which include the configuration file).

        Args:
            cli_args (list): arguments of the command line interface

        Returns:
            str, str, str: init_config, aas_model and ontology values (None if not specified).
        """
        parser = argparse.ArgumentParser(description='Parser for SMIA CLI arguments')
        parser.add_argument("-m", "--model")
        # parser.add_argument("-o", "--ontology")
        parser.add_argument("-c", "--config")
        args = parser.parse_args(cli_args)
        return args.config, args.model

    @staticmethod
    def check_and_save_cli_information(init_config, aas_model):
        """
        This method checks the information added in the CLI and, depending on the result of the check, it gets the
        necessary information and saves it in the appropriate variable.

        Args:
            init_config (str): path to the initialization configuration properties file.
            aas_model (str): path to the AAS model file.
        """
        # The libraries are imported locally to avoid circular import error
        from smia.aas_model.aas_model_utils import AASModelUtils
        from smia.utilities import smia_archive_utils
        from smia.utilities import properties_file_utils
        import ntpath

        if (init_config is None) and (aas_model is None):
            # If all are none, the CLI information is invalid
            raise CriticalError("The CLI information is invalid: no argument has been specified. Please specify at "
                                "least one")
        elif init_config is not None:
            # If the configuration properties file is defined, its variable is updated
            init_config_file_name = ntpath.split(init_config)[1] or ntpath.basename(ntpath.split(init_config)[0])
            SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME = init_config_file_name

            if aas_model is None:
                # If the configuration file is specified, but not the AAS model, the AAS must be defined in the file
                # and obtained from there.
                aas_model_folder_path = properties_file_utils.get_aas_general_property('model.folder')
                aas_model_file_name = properties_file_utils.get_aas_general_property('model.file')
                if (aas_model_file_name is None) or (aas_model_folder_path is None):
                    raise CriticalError("The CLI information is invalid: the AAS model has not been specified either "
                                        "in the CLI or in the configuration properties file.")

                smia_archive_utils.copy_file_into_archive(aas_model_folder_path + '/' + aas_model_file_name,
                                                          SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH)
                GeneralUtils.update_aas_model(aas_model_file_name)
                # SMIAGeneralInfo.CM_AAS_MODEL_FILENAME = aas_model_file_name

        elif aas_model is not None:
            # If the AAS model file is defined, its variable is updated
            aas_model_file_name = ntpath.split(aas_model)[1] or ntpath.basename(ntpath.split(aas_model)[0])
            GeneralUtils.update_aas_model(aas_model_file_name)
            # SMIAGeneralInfo.CM_AAS_MODEL_FILENAME = aas_model_file_name

            if init_config is None:
                # If the configuration properties file is not defined, it is obtained from inside the AASX package
                if aas_model_file_name.split(".")[1] != 'aasx':
                    # If it is not an AASX package, the configuration file is not specified
                    raise CriticalError("The configuration file is not specified and is not inside the AAS model as it "
                                        "is not an AASX package. Please specify the configuration file or add it to "
                                        "an AASX package and pass it as an AAS model.")
                config_file_path = AASModelUtils.get_configuration_file_path_from_standard_submodel()
                # smia_archive_utils.copy_file_into_archive(config_file_path,
                #                                           SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH)
                init_config_file_name = ntpath.split(config_file_path)[1] or ntpath.basename(ntpath.split(config_file_path)[0])
                # The file must be obtained from the AASX package
                config_file_bytes = AASModelUtils.get_file_bytes_from_aasx_by_path(config_file_path)
                if config_file_bytes is None:
                    CriticalError("The CLI information is invalid: the initialization configuration file has not "
                                    "been specified either in the CLI or inside the AAS model.")
                with open(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + init_config_file_name,
                          "wb") as binary_file:
                    # Write bytes to file
                    binary_file.write(config_file_bytes)
                SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME = init_config_file_name

_logger = logging.getLogger(__name__)

class DockerUtils:
    """This class contains utility methods related to SMIA execution in Docker containers."""

    @staticmethod
    def get_aas_model_from_env_var():
        """
        This method returns the AAS model path, obtained from the required 'AAS_MODEL_NAME' environmental variable.

        Returns:
            str: path to the AAS model to be loaded.
        """
        aas_model_name = os.environ.get('AAS_MODEL_NAME')
        if aas_model_name is None:
            _logger.error("The environment variable 'AAS_MODEL_NAME' for the AAS model is not set, so SMIA cannot start. "
                          "Please add the information and restart the container.")
            return
        _logger.info('Loaded AAS model: {}'.format(aas_model_name))
        aas_model_path = SMIAGeneralInfo.CONFIGURATION_AAS_FOLDER_PATH + '/' + aas_model_name
        return aas_model_path



class ExtendedAssetAdministrationShell:
    """This class contains methods to be added to AssetAdministrationShell class of Basyx Python SDK model."""

    def print_aas_information(self):
        print("AAS information:")
        print("\tid: " + self.id)
        print("\tid_short: " + self.id_short)
        print("\tdisplayName:{}".format(self.display_name))
        print("\tdescription:{}".format(self.description))
        print("\tcategory: {}".format(self.category))
        print("\tderivedFrom: {}".format(self.derived_from))
        print("\tadministration: {}".format(ExtendedGeneralMethods.print_administration(self.administration)))
        print("\textension: " + "{}".format(ExtendedGeneralMethods.print_namespace_set(self.extension)))
        print("\tdataSpecifications: " + "{}".format(
            ExtendedGeneralMethods.print_data_specifications(self.embedded_data_specifications)))


class ExtendedAssetInformation:
    """This class contains methods to be added to AssetInformation class of Basyx Python SDK model."""

    def print_asset_information(self):
        print("Asset information:")
        print("\tassetKind: {}".format(self.asset_kind))
        print("\tassetType: {}".format(self.asset_type))
        print("\tspecificAssetId: {}".format(self.specific_asset_id))
        print("\tglobalAssetId: {}".format(self.global_asset_id))
        print("\tdefaultThumbnail: {}".format(self.default_thumbnail))


class ExtendedGeneralMethods:

    @staticmethod
    def print_administration(administration):
        if administration:
            return "version[{}".format(administration.version) + "], revision[{}".format(
                administration.revision) + "], creator[{}".format(administration.creator) + "], templateId[{}".format(
                administration.template_id) + "]"
        else:
            return ""

    @staticmethod
    def print_namespace_set(namespace_set):
        string = ""
        for item in namespace_set:
            string += str(item) + ","
        return string

    @staticmethod
    def print_data_specifications(embedded_data_specifications):
        string = ""
        for item in embedded_data_specifications:
            string += ("(Reference: {}".format(item.data_specification) +
                       " | Content: {}".format(item.data_specification_content) + "),")
        return string



"""
This is the SMIA launch file for CLI, which executes the program logic from the command and gets the AAS model from the
arguments.
"""

import logging
import os
import sys

# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))    # To run in CLI (executing in 'src' folder)

import smia
from smia.agents.smia_agent import SMIAAgent
from smia.agents.smia_app_agent import SMIAAppAgent
from smia.agents.smia_resource_agent import SMIAResourceAgent
from smia.utilities import smia_archive_utils
from smia.utilities import properties_file_utils
from smia.utilities.general_utils import CLIUtils

# XMPP_SERVER = 'worker4'
# XMPP_SERVER = 'ejabberd'

_logger = logging.getLogger(__name__)

def main():
    # First, the command line arguments are obtained
    init_config, aas_model = CLIUtils.get_information_from_cli(sys.argv[1:])
    smia_archive_utils.save_cli_added_files(init_config, aas_model)

    # It is checked and saved the command line arguments
    CLIUtils.check_and_save_cli_information(init_config, aas_model)

    # The AAS_ID will be set in the associated ConfigMap, within the general-information of the AAS
    aas_id = properties_file_utils.get_dt_general_property('agentID')
    passwd = properties_file_utils.get_dt_general_property('password')

    # The XMPP server of the MAS will also be set in the associated ConfiMap
    xmpp_server = properties_file_utils.get_dt_general_property('xmpp-server')

    # Build the agent jid and password
    agent_jid = aas_id + '@' + xmpp_server

    # Depending on the asset type, the associated SPADE agent will be created
    aas_type = ''  # For testing (TODO Get the type of the asset)
    match aas_type:
        case "physical":
            _logger.info("The asset is physical")
            smia_agent = SMIAResourceAgent(agent_jid, passwd)
        case "digital":
            _logger.info("The asset is logical")
            smia_agent = SMIAAppAgent(agent_jid, passwd)
        case _:
            _logger.info("The asset is not defined, so it is a generic SMIA")
            # Create the agent object
            smia_agent = SMIAAgent(agent_jid, passwd)

    smia.run(smia_agent)


if __name__ == '__main__':
    # First, the initial configuration must be executed
    smia.initial_self_configuration()
    _logger.info("Initializing SMIA software...")

    # Run main program with SPADE
    main()



import asyncio
import logging
import sys

import basyx.aas.adapter.json
import basyx.aas.adapter.xml
from basyx.aas import model
from basyx.aas.model import ModelReference
from spade.behaviour import OneShotBehaviour
from tqdm.asyncio import tqdm

from smia.aas_model import extended_submodel
from smia.aas_model.aas_model_utils import AASModelUtils
from smia.aas_model.extended_submodel import ExtendedSkill, ExtendedSkillInterface, ExtendedComplexSkillInterface, \
    ExtendedComplexSkill, ExtendedSimpleSkill, ExtendedSimpleSkillInterface
from smia.assetconnection.http_asset_connection import HTTPAssetConnection
from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyUtils, CapabilitySkillOntologyInfo, \
    CSSModelAASModelInfo
from smia.utilities.smia_info import AssetInterfacesInfo
from smia.logic.exceptions import AASModelReadingError, AASModelOntologyError, \
    OntologyReadingError, OntologyInstanceCreationError

_logger = logging.getLogger(__name__)


class InitAASModelBehaviour(OneShotBehaviour):
    """
    This class implements the behaviour responsible for reading the AAS model to obtain all submodels, submodel elements
     and concept descriptions. The necessary checks are performed to ensure the valid  initial conditions to start the
     running state of the DT. The AAS model is defined in the ConfigMap associated to the deployed container and in
     order to successfully read the definition in any serialization format (JSON or XML), BaSyx Python SDK will be used.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

        self.progress_bar = None

        # The counters for all type of analyzed elements are initialized
        self.analyzed_capabilities = []
        self.analyzed_capability_constraints = []
        self.analyzed_skills = []
        self.analyzed_skill_interfaces = []
        self.analyzed_skill_params = []
        self.analyzed_asset_connections = []
        self.errors_found = []

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """
        # Depending on the serialization format, the required BaSyx read method shall be executed. This will store all
        # the received elements of the model in the corresponding global object of the agent.
        object_store = AASModelUtils.read_aas_model_object_store()
        await self.myagent.aas_model.set_aas_model_object_store(object_store)

        # First, the initial general analysis is performed on the AAS model.
        await self.myagent.aas_model.execute_general_analysis(self.myagent)

        # A progress bar is used for showing how the AAS model is being read.
        await self.create_progress_bar_object()

        # When the object store is created, the required values and information is obtained from the AAS model
        _logger.info("Reading the AAS model to get all defined ontology elements...")
        await self.get_and_save_ontology_classes_information()

        await asyncio.sleep(.5)
        _logger.info("Reading the AAS model to get all relationships between ontology elements...")
        await self.get_and_save_ontology_relationships_information()

        # After the AAS model has been analyzed, the AssetConnection class can be specified
        _logger.info("Reading the AAS model to get all asset connections...")
        await self.get_and_configure_asset_connections()

        # The progress bar is closed
        self.progress_bar.close()
        await asyncio.sleep(1)

        # The final results of the analysis are shown
        await self.print_analysis_result()

        _logger.info("AAS model initialized.")
        self.exit_code = 0

    async def get_and_save_ontology_classes_information(self):
        """
        This method stores all the information related to the class elements defined in the ontology. Since the data is
         defined in the AAS model, it will be used to check whether the required data defined in the ontology has been
          added in the AAS model. If the elements are valid, they will be created their associated ontology instance and
          the AAS SubmodelElement will be associated to these instances.
        """
        for ontology_class_iri in CapabilitySkillOntologyInfo.CSS_ONTOLOGY_THING_CLASSES_IRIS:
            await self.check_and_create_instances_by_iri(ontology_class_iri)

    async def get_and_save_ontology_relationships_information(self):
        """
        This method stores all the information related to the relationships between elements defined in the ontology.
        Since the data is defined in the AAS model, it will be used to check whether the linked elements have their
        associated ontology instance (created just before the execution of this method).
        """
        for ontology_relationship_iri in CapabilitySkillOntologyInfo.CSS_ONTOLOGY_OBJECT_PROPERTIES_IRIS:
            await self.check_and_create_relationship_by_iri(ontology_relationship_iri)

        # Skills that are of type Operation SubmodelElement are a special case, because their OperationVariables are
        # related SkillParameters, but there is no relation between them.
        await self.check_and_create_operation_relationships()

    async def check_and_create_instances_by_iri(self, ontology_class_iri):
        """
        This method checks the relationship between two elements and, if it is valid, it creates it (it connects the
        related ontology instances through the appropriate ObjectProperty).

        Args:
            ontology_class_iri (str): identifier in form of IRI for the element within the CSS ontology.
        """
        sme_list = None
        try:
            sme_list = await self.myagent.aas_model.get_submodel_elements_by_semantic_id(ontology_class_iri)
        except (AASModelReadingError, AASModelOntologyError, OntologyReadingError) as e:
            if isinstance(e, AASModelReadingError) or isinstance(e, AASModelOntologyError):
                _logger.warning("Check the AAS Model {}. Reason of the fail: {}.".format(e.sme_class, e.reason))
            else:
                _logger.warning("Check the CSS ontology definition.")

        for submodel_elem in sme_list:
            await self.add_step_progress_bar()

            try:
                # For each SubmodelElement an instance within the CSS ontology is created
                await self.create_ontology_instance_from_sme_element(submodel_elem, ontology_class_iri)
                # created_instance = await self.myagent.css_ontology.create_ontology_object_instance(ontology_class,
                #                                                                                    submodel_elem.id_short)

                # The submodel element will be converted from the Basyx class to the extended one (of the SMIA approach)
                await self.convert_sme_class_to_extended_by_iri(submodel_elem, ontology_class_iri)

                # The element is saved as analyzed
                await self.add_new_analyzed_element(submodel_elem)

            except (AASModelReadingError, AASModelOntologyError, OntologyReadingError) as e:
                if isinstance(e, AASModelReadingError) or isinstance(e, AASModelOntologyError):
                    _logger.warning("Check the AAS Model {}. Reason of the fail: {}.".format(e.sme_class, e.reason))
                else:
                    _logger.warning("Check the CSS ontology definition.")
                self.errors_found.append(submodel_elem.id_short)

    async def create_ontology_instance_from_sme_element(self, sme_elem, ontology_iri):
        """
        This method creates the ontology instance from the AAS Submodel Element.

        Args:
            sme_elem (basyx.aas.model.SubmodelElement): SubmodelElement of the AAS model with all configured data.
            ontology_iri (str): IRI of the ontology class of the instance to be created.
        """
        try:
            ontology_class = await self.myagent.css_ontology.get_ontology_class_by_iri(ontology_iri)
            if ontology_class is None:
                raise OntologyReadingError('The ontology class with IRI {} does not exist in the given OWL ontology. '
                                           'Check the ontology file.')
            # The class is used as constructor to build the instance of the CSS ontology
            created_instance = await self.myagent.css_ontology.create_ontology_object_instance(ontology_class,
                                                                                               sme_elem.id_short)
            # TODO de momento se crean las instancias de las ontologias con el id_short, pero habria que pensar si el
            #  id_short puede repetirse (el id_short es único dentro del submodelo)
            await self.add_ontology_required_information(sme_elem, created_instance)

        except (AASModelReadingError, AASModelOntologyError, OntologyReadingError, OntologyInstanceCreationError) as e:
            if isinstance(e, AASModelReadingError) or isinstance(e, AASModelOntologyError):
                _logger.warning("Check the AAS Model {}. Reason of the fail: {}.".format(e.sme_class, e.reason))
            else:
                _logger.warning("Check the CSS ontology definition.")
            self.errors_found.append(sme_elem.id_short)

    @staticmethod
    async def add_ontology_required_information(aas_model_elem, ontology_instance):
        """
        This method adds the required information defined in the ontology to the given instance. All information is
        obtained from given the AAS Submodel Element.

        Args:
            aas_model_elem (basyx.aas.model.SubmodelElement): SubmodelElement of the AAS model with all configured data.
            ontology_instance (ThingClass): instance class on which the information will be added.
        """
        # The ontology may state that it is required to add some attributes
        ontology_required_value_iris = ontology_instance.get_data_properties_iris()
        for required_value_iri in ontology_required_value_iris:
            required_value = aas_model_elem.get_qualifier_value_by_semantic_id(required_value_iri)
            required_value_name = ontology_instance.get_data_property_name_by_iri(required_value_iri)
            ontology_instance.set_data_property_value(required_value_name, required_value)
        # The Submodel Element is also added to be available to the ontology instance object in form of a reference
        ontology_instance.set_aas_sme_ref(ModelReference.from_referable(aas_model_elem))

    @staticmethod
    async def convert_sme_class_to_extended_by_iri(sme_elem, ontology_iri):
        """
        This method converts the class of a SubmodelElement to the Extended class, in order to add the required method
        to be used during the execution of the software. The extended class is obtained from from the CSS ontology utils
         class using the IRI of the ontology class.

        Args:
            sme_elem (basyx.aas.model.SubmodelElement): SubmodelElement of the AAS model to be modified.
            ontology_iri (iri): ontology class IRI.
        """
        current_class = sme_elem.__class__
        new_class = CSSModelAASModelInfo.CSS_ONTOLOGY_AAS_MODEL_LINK[ontology_iri]
        # TODO comprobar si con el nuevo metodo de extension en 'add_old_sme_class' se evita el problema del Simple y Complex de skill e interfaz
        if new_class is ExtendedSkill:
            # In this case there are two types of classes (Simple and Complex)
            if sme_elem.check_if_element_is_structural():
                sme_elem.__class__ = ExtendedComplexSkill
            else:
                sme_elem.__class__ = ExtendedSimpleSkill
        elif new_class is ExtendedSkillInterface:
            # In this case there are two types of classes (Simple and Complex)
            if sme_elem.check_if_element_is_structural():
                sme_elem.__class__ = ExtendedComplexSkillInterface
            else:
                sme_elem.__class__ = ExtendedSimpleSkillInterface
        else:
            sme_elem.__class__ = new_class
        # The old class new to be added to the Extended class
        sme_elem.add_old_sme_class(current_class)

    async def check_and_create_relationship_by_iri(self, relationship_iri):
        """
        This method checks the relationship between two elements and, if it is valid, it creates it (it connects the
        related ontology instances through the appropriate ObjectProperty).

        Args:
            relationship_iri (str): identifier in form of IRI for the relationship within the CSS ontology.
        """
        rels_list, rel_ontology_class, domain_aas_class, range_aas_class = None, None, None, None
        try:
            rels_list = await self.myagent.aas_model.get_submodel_elements_by_semantic_id(
                relationship_iri, basyx.aas.model.RelationshipElement)
            rel_ontology_class = await self.myagent.css_ontology.get_ontology_class_by_iri(relationship_iri)
            # The required AAS classes for the elements within this relationship is obtained from the CSS ontology
            domain_aas_class, range_aas_class = CapabilitySkillOntologyUtils.get_aas_classes_from_object_property(
                rel_ontology_class)
        except (AASModelReadingError, AASModelOntologyError, OntologyReadingError) as e:
            if isinstance(e, AASModelReadingError) or isinstance(e, AASModelOntologyError):
                _logger.warning("Check the AAS Model {}. Reason of the fail: {}.".format(e.sme_class, e.reason))
            else:
                _logger.warning("Check the CSS ontology definition.")

        domain_aas_elem, range_aas_elem = None, None
        for rel in rels_list:
            await self.add_step_progress_bar()

            # First, the elements of capability and skill are determined (no matter in which order of the
            # relationship they are listed).
            try:
                domain_aas_elem, range_aas_elem = await self.myagent.aas_model.get_elements_from_relationship(
                    rel, domain_aas_class, range_aas_class)
                # It is checked if the capability and the skill have the required semanticIDs within the ontology
                # if not domain_aas_elem.check_semantic_id_exist(domain_class_iri):
                domain_aas_elem.get_semantic_id_of_css_ontology()
                range_aas_elem.get_semantic_id_of_css_ontology()

                # When both are checked, if the ontology instances exist the link is set in the ontology
                await self.myagent.css_ontology.add_object_property_to_instances_by_names(rel_ontology_class.name,
                                                                                          domain_aas_elem.id_short,
                                                                                          range_aas_elem.id_short)

            except (AASModelReadingError, AASModelOntologyError, OntologyReadingError) as e:
                if isinstance(e, AASModelReadingError) or isinstance(e, AASModelOntologyError):
                    _logger.warning("Check the AAS Model {}. Reason of the fail: {}.".format(e.sme_class, e.reason))
                else:
                    _logger.warning("Check the CSS ontology definition.")
                if None not in (domain_aas_elem, range_aas_elem):
                    self.errors_found.append("{} ({},{})".format(rel.id_short, domain_aas_elem.id_short,
                                                                 range_aas_elem.id_short))
                else:
                    self.errors_found.append("{} ({},{})".format(rel.id_short, domain_aas_class, range_aas_class))

    async def check_and_create_operation_relationships(self):
        """
        This method checks the relationship between OperationVariables and their related Operation SubmodelElement and,
        if they are included in the CSS ontology (by semanticIDs), the related ontology instances are linked through
        the appropriate ObjectProperty.
        """
        skill_list = await self.myagent.aas_model.get_submodel_elements_by_semantic_id(
            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI)
        rel_ontology_class = await self.myagent.css_ontology.get_ontology_class_by_iri(
            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_PROP_HASPARAMETER_IRI)
        for skill in skill_list:
            if isinstance(skill, model.Operation):
                variable = None
                try:
                    if rel_ontology_class is None:
                        raise OntologyReadingError('The class for the relationship with IRI {} does not exist in the '
                                                   'given ontology. Check the OWL ontology file.'.format(
                            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_PROP_HASPARAMETER_IRI))
                    if skill.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI):
                        operation_variables = skill.get_operation_variables_by_semantic_id(
                            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_PARAMETER_IRI)
                        for variable in operation_variables:
                            # When both are checked, if the ontology instances exist the link is set in the ontology
                            await self.myagent.css_ontology.add_object_property_to_instances_by_names(
                                rel_ontology_class.name, skill.id_short, variable.id_short)
                except OntologyReadingError as error:
                    _logger.warning("Check the CSS ontology definition.")
                    if None not in (skill, variable):
                        self.errors_found.append("{} ({},{})".format('hasParameter', skill.id_short,
                                                                     variable.id_short))
                    else:
                        self.errors_found.append("{} ({},{})".format('hasParameter', 'Operation', 'OperationVariable'))


    async def get_and_configure_asset_connections(self):
        """
        This method gets all Asset Interfaces Descriptions in the AAS model, configures each case with associated
        'AssetConnection' class and saves the information in the global variable of the agent for all asset connections.
        """
        # TODO PROXIMO PASO: hay que modificicar la variable global para assetconnection del agente ya que se debe
        #  habilitar añadir mas de un AssetConnection. Para ello, se analizara el Submodelo "AssetInterfacesDescription"
        #  y se recogerá cada SMC dentro de el, el cual determinara una interfaz diferente. Por cada interfaz,
        #  dependiendo del tipo (HTTP, OPC UA...) se generará una clase AssetConnection. Esta deberá tener, por una
        #  parte, la referencia al SMC del modelo AAS. Utilizando esta referencia, leera toda la informacion de ese
        #  submodelo y configurará la interfaz (p.e. añadiendo los datos del endpoint), para dejar la interfaz lista
        #  para ser usada (se hará referencia a los SMEs dentro de "InteractionMetada" del submodelo de la interfaz,
        #  que es donde estan los datapoints)
        asset_interfaces_submodel = await self.myagent.aas_model.get_submodel_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL)
        if not asset_interfaces_submodel:
            _logger.warning("AssetInterfacesSubmodel submodel is not defined. Make sure that this DT does not need to "
                            "be connected to the asset.")
            return
        for interface_elem in asset_interfaces_submodel.submodel_element:
            await self.add_step_progress_bar()

            if interface_elem.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACE) is False:
                _logger.warning("There is a submodel element inside the interfaces submodel with invalid semanticID.")
                continue
            # Dependiendo del tipo de interfaz se generara una clase u otra (de momento solo HTTP)
            if interface_elem.check_suppl_semantic_id_exist(AssetInterfacesInfo.SUPPL_SEMANTICID_HTTP):
                # Hay una interfaz de tipo HTTP
                http_connection_class = HTTPAssetConnection()
                await http_connection_class.configure_connection_by_aas_model(interface_elem)
                interface_model_ref = ModelReference.from_referable(interface_elem)
                await self.myagent.save_asset_connection_class(interface_model_ref, http_connection_class)
            if interface_elem.check_suppl_semantic_id_exist('id de opc ua'):
                # TODO Hay una interfaz de tipo OPC CUA
                pass

            self.analyzed_asset_connections.append(interface_elem.id_short)

        _logger.info("All asset connections defined in the AAS model have been configured and saved.")

    async def print_analysis_result(self):
        """
        This method simply prints the result of the complete analysis of the AAS model.
        """
        _logger.info("AAS model analysis results. \n\t- Analyzed capabilities: {}\n\t- Analyzed capability "
                     "constraints: {}\n\t- Analyzed skills: {}\n\t- Analyzed skill interfaces: {}\n\t- Analyzed skill "
                     "parameters: {}\n\t- Analyzed asset connections: {}\n\t\x1b[91m- Errors found: {}\x1b[0m".format(
                      self.analyzed_capabilities, self.analyzed_capability_constraints, self.analyzed_skills,
                      self.analyzed_skill_interfaces, self.analyzed_skill_params, self.analyzed_asset_connections,
                      self.errors_found))

        for analyzed_dict in [self.analyzed_capabilities, self.analyzed_capability_constraints, self.analyzed_skills,
                     self.analyzed_skill_interfaces, self.analyzed_skill_params, self.analyzed_asset_connections]:
            if len(analyzed_dict) > 0 and set(analyzed_dict).issubset(self.errors_found):
                _logger.error("All elements of an ontology class have been detected as errors, so check the AAS model "
                              "or the OWL ontology file.")

    async def create_progress_bar_object(self):
        """
        This method creates the object for showing by console the progress of the analysis of the AAS model in form of
        a progress bar. The object is
        """
        # The iterations during the analysis of the AAS model will be the number of steps in the progress bar
        # The iteration number will be the sum of the possible Asset Connections and the Capability-Skill relationships.
        ontology_elements_semantic_ids = CapabilitySkillOntologyInfo.CSS_ONTOLOGY_THING_CLASSES_IRIS + \
                                         CapabilitySkillOntologyInfo.CSS_ONTOLOGY_OBJECT_PROPERTIES_IRIS
        ontology_elements_list = await self.myagent.aas_model.get_submodel_elements_by_semantic_id_list(
            ontology_elements_semantic_ids)
        asset_interfaces_submodel = await self.myagent.aas_model.get_submodel_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL)
        if not asset_interfaces_submodel:
            _logger.warning(
                "AssetInterfacesSubmodel submodel is not defined. Make sure that this DT does not need to be"
                " connected to the asset.")
            asset_interfaces_submodel = type('obj', (object,), {
                'submodel_element': []})  # This is a solution to not brake the next commmand
        total_iterations = len(asset_interfaces_submodel.submodel_element) + len(ontology_elements_list)
        # with logging_redirect_tqdm():
        self.progress_bar = tqdm(total=total_iterations, desc='Analyzing AAS model', file=sys.stdout, ncols=75,
                                 bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} ontology submodel elements \n')

    async def add_step_progress_bar(self):
        """
        This method adds a new step in the progress bar to show the status of the AAS model analysis.
        """
        # Add a new step in the progress bar
        self.progress_bar.update(1)
        await asyncio.sleep(.2)  # Simulate some processing time

    async def add_new_analyzed_element(self, aas_elem):
        """
        This method adds the new analyzed element in the corresponding list of class variables.

        Args:
            aas_elem (basyx.aas.model.SubmodelElement): analyzed SubmodelElement,
        """
        if isinstance(aas_elem, extended_submodel.ExtendedCapability):
            self.analyzed_capabilities.append(aas_elem.id_short)
        elif isinstance(aas_elem, extended_submodel.ExtendedSkill):
            self.analyzed_skills.append(aas_elem.id_short)
        elif isinstance(aas_elem, extended_submodel.ExtendedSkillInterface):
            self.analyzed_skill_interfaces.append(aas_elem.id_short)
        elif isinstance(aas_elem, extended_submodel.ExtendedSkillParameter):
            self.analyzed_skill_params.append(aas_elem.id_short)
        elif isinstance(aas_elem, extended_submodel.ExtendedCapabilityConstraint):
            self.analyzed_capability_constraints.append(aas_elem.id_short)



# """This module contains all OWL clases in form of Python classes, in order to add required methods to the correct
# execution of SMIA software. This module is associated to the OWL ontology since it is defined inside the file of the
# definition of OWL. When the owlready2 package loads the ontology, it will automatically import this Python module."""
import logging
from itertools import chain

import basyx.aas.model
from owlready2 import Thing, get_ontology, DataPropertyClass, DatatypeClass

from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyInfo, CapabilitySkillOntologyUtils, \
    CSSModelAASModelInfo
from smia.logic.exceptions import OntologyCheckingAttributeError, OntologyCheckingPropertyError, \
    OntologyInstanceCreationError

_logger = logging.getLogger(__name__)

import builtins
if hasattr(builtins, '__sphinx_build__'):
    print("Sphinx build is running, so, to correctly import this module the ontology and namespace must be initialized.")
    css_ontology = None  # It is necessary to build Sphinx documentation without errors.
    base_namespace = None   # It is necessary to build Sphinx documentation without errors.
else:
    css_ontology = get_ontology(CapabilitySkillOntologyUtils.get_ontology_file_path())
    base_namespace = css_ontology.get_namespace(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_BASE_NAMESPACE)

# css_ontology = get_ontology(CapabilitySkillOntologyUtils.get_ontology_file_path())
# base_namespace = css_ontology.get_namespace(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_BASE_NAMESPACE)
# css_ontology = None   # It is necessary to build Sphinx documentation without errors.
# base_namespace = None   # It is necessary to build Sphinx documentation without errors.


class ExtendedThing(Thing):

    # The namespace of the base CSS ontology must be defined
    namespace = base_namespace

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Some dictionaries related to the Data Properties are initialized
        self.data_properties_dict = set()
        self.data_properties_types_dict = {}
        self.data_properties_values_dict = {}

        # The reference to the associated AAS model element will be also necessary.
        self.aas_sme_ref = None

        # The data properties associated to this instance class are found
        self.seek_associated_data_properties()

    @classmethod
    def get_associated_aas_class(cls):
        """
        This method gets the associated AAS model class of this ontology class. The associated AAS model class will be
        obtained from the CSS ontology utils class using the IRI of the class.

        Returns:
            basyx.aas.model.SubmodelElement: AAS model class.
        """
        if cls.iri not in CSSModelAASModelInfo.CSS_ONTOLOGY_AAS_MODEL_LINK:
            # If this ontology class does not have a related AAS model class, the general SubmodelElement class is set
            return basyx.aas.model.SubmodelElement
        else:
            return CSSModelAASModelInfo.CSS_ONTOLOGY_AAS_MODEL_LINK[cls.iri]

    def seek_associated_data_properties(self):
        """
        This method seeks possible limited values for attributes of Capability, in order to validate when the attribute
        value is assigned. The possible values for limited attributes are stored in a global dictionary.
        """
        for prop in css_ontology.properties():
            if isinstance(prop, DataPropertyClass):
                if CapabilitySkillOntologyUtils.check_whether_part_of_domain(self, prop.domain):
                    # First, the data property object is saved
                    self.data_properties_dict.add(prop)
                    # The possible values of the data properties are also stored
                    for range_value in prop.range:
                        if isinstance(range_value, DatatypeClass):
                            possible_values = CapabilitySkillOntologyUtils.get_possible_values_of_datatype(range_value)
                            xsd_value_type = CapabilitySkillOntologyUtils.check_and_get_xsd_datatypes(range_value)
                            if possible_values is not None:
                                self.data_properties_types_dict[prop.name] = possible_values
                            elif xsd_value_type is not None:
                                self.data_properties_types_dict[prop.name] = xsd_value_type
                        else:
                            self.data_properties_types_dict[prop.name] = range_value

    def check_valid_data_property_value(self, data_property_name, data_property_value):
        """
        This method checks if the given value of a given data property is valid, in terms of the type of the data. If
        the data property type is simple, the type will be directly checked, and in case of an enumeration, if the given
        value is within the possible values will be checked.

        Args:
            data_property_name (str): the name of the data property.
            data_property_value (str): the value of the data property to be checked.
        """
        if data_property_name not in self.data_properties_types_dict:
            _logger.warning("The data property {} does not exist in this OWL class ({}).".format(data_property_name,
                                                                                                 self))
        else:
            data_property_type = self.data_properties_types_dict[data_property_name]
            if isinstance(data_property_type, set):
                if data_property_value not in data_property_type:
                    raise OntologyInstanceCreationError("The data property value {} for the OWL class {} is not within "
                                                        "the valid values.".format(data_property_value, self))
            else:
                if not isinstance(data_property_value, data_property_type):
                    raise OntologyInstanceCreationError("The data property value {} for the OWL class {} is not valid."
                                                        .format(data_property_value, self))

    def set_data_property_value(self, data_property_name, data_property_value):
        """
        This method sets the value of a data property associated to the self instance only if the given value is within
         the possible values for this attribute.

        Args:
            data_property_name (str): the name of the data property.
            data_property_value (str): the value of the data property to be stored.
        """
        self.check_valid_data_property_value(data_property_name, data_property_value)
        # If the result of the check is True the execution reaches this point
        self.data_properties_values_dict[data_property_name] = data_property_value

    def get_data_properties_names(self):
        """
        This method gets all names of the data properties associated to the self instance class (obtained during the
        initialization of the class).

        Returns:
            list(str): list with all names of the data properties.
        """
        data_properties_iris = []
        for prop in self.data_properties_dict:
            data_properties_iris.append(prop.name)
        return data_properties_iris

    def get_data_properties_iris(self):
        """
        This method gets all IRIs of the data properties associated to the self instance class (obtained during the
        initialization of the class).

        Returns:
            list(str): list with all IRIs of the data properties.
        """
        data_properties_iris = []
        for prop in self.data_properties_dict:
            data_properties_iris.append(prop.iri)
        return data_properties_iris

    def get_data_property_name_by_iri(self, property_iri):
        """
        This method gets the name of the data property associated to the self instance class (obtained during the
        initialization of the class). It is found by its IRI.

        Args:
            property_iri (str): IRI of the data property to find.

        Returns:
            str: name of the desired data property.
        """
        for prop in self.data_properties_dict:
            if prop.iri == property_iri:
                return prop.name
        _logger.warning("The data property with IRI {} does not exist in class {}".format(property_iri, self))
        return None

    def get_aas_sme_ref(self):
        """
        This method gets the AAS submodel element (SME) related to this instance class.

        Returns:
            aas_ref (str): reference to the submodel element of the AAS model.
        """
        return self.aas_sme_ref

    def set_aas_sme_ref(self, aas_ref):
        """
        This method sets the AAS submodel element (SME) related to this instance class.

        Args:
            aas_ref (str): reference to the submodel element of the AAS model.
        """
        self.aas_sme_ref = aas_ref

    def check_and_get_related_instance_by_instance_name(self, other_instance_name):
        """
        This method checks if there is some Object Property defined that connects the self instance class with the given
        instance class,i.e., if the instances are related within the ontology. If the relation is valid, it also
        returns the related ThingClass.

        Args:
            other_instance_name (ThingClass): name of the other instance class.

        Returns:
            bool, ThingClass: the result of the check, and if True, the class of the related instance
        """
        for prop in self.get_properties():
            for related_instance in getattr(self, prop.name):
                if other_instance_name == related_instance.name:
                    return True, related_instance
        return False, None


class Capability(ExtendedThing):
    """
    This class represent the OWL class for Capabilities. It contains all necessary methods to ensure the correct
    execution of SMIA software.
    """

    # The associated SubmodelElement class of the AAS is also defined
    # _aas_sme_class = 1
    # aas_sme_class = extended_submodel.ExtendedCapability
    # aas_sme_class = basyx.aas.model.Capability


    def check_instance(self):
        """
        This method checks whether the Capability instance is valid: if the required attributes are set and if all the
        added properties are valid. In case of invalid Capability, it raises the exception related to the checking error.
        """
        if self.has_lifecycle is None:
            raise OntologyCheckingAttributeError("The 'has_lifecycle' attribute is required in "
                                                 "Capability instances.", self)
        for skill in self.isRealizedBy:
            if not isinstance(skill, Skill):
                raise OntologyCheckingPropertyError("The instance {} is added in 'isRealizedBy' and it is not a "
                                                    "Skill".format(skill), 'isRealizedBy', skill)
        for constraint in self.isRestrictedBy:
            if not isinstance(constraint, CapabilityConstraint):
                raise OntologyCheckingPropertyError("The instance {} is added in 'isRestrictedBy' and it is not a "
                                                    "CapabilityConstraint".format(constraint), 'isRestrictedBy',
                                                    constraint)
        # TODO añadir validacion de que tiene una referencia de AAS valida
        # TODO pensar mas tipos de validaciones

    def get_associated_skill_instances(self):
        """
        This method gets all associated skill instances and, if there is no skill, returns the None object.

        Returns:
            IndividualValueList: generator with all associated skill instances.
        """
        if len(self.isRealizedBy) == 0:
            return None
        else:
            return self.isRealizedBy

    def get_associated_constraint_instances(self):
        """
        This method gets all associated constraint instances and, if there is no one, returns the None object.

        Returns:
            IndividualValueList: generator with all associated constraint instances.
        """
        if len(self.isRestrictedBy) == 0:
            return None
        else:
            return self.isRestrictedBy

class CapabilityConstraint(ExtendedThing):

    # The associated SubmodelElement class of the AAS is also defined
    # aas_sme_class = None
    # aas_sme_class = basyx.aas.model.SubmodelElement
    # aas_sme_class = extended_submodel.ExtendedCapabilityConstraint

    # TODO PENSAR METODOS PARA CONSTRAINTS
    pass


class Skill(ExtendedThing):

    # The associated SubmodelElement class of the AAS is also defined
    # aas_sme_class = None
    # aas_sme_class = extended_submodel.ExtendedSkill

    def check_instance(self):
        """
        This method checks whether the Skill instance is valid: if the required attributes are set and if all the added
         properties are valid. In case of invalid Skill, it raises the exception related to the checking error.
        """
        if len(self.accessibleThrough) == 0:
            raise OntologyCheckingAttributeError(
                "The instance {} does not have any SkillInterface associated".format(self), self)
        # TODO pensar mas comprobaciones

    def get_associated_skill_interface_instances(self):
        """
        This method gets all associated skill interface instances and, if there is no one, returns the None object.

        Returns:
            IndividualValueList: generator with all associated skill interface instances.
        """
        if len(self.accessibleThrough) + len(self.accessibleThroughAgentService) + len(self.accessibleThroughAssetService) == 0:
            return None
        else:
            return chain(self.accessibleThrough, self.accessibleThroughAgentService, self.accessibleThroughAssetService)

    def get_associated_skill_parameter_instances(self):
        """
        This method gets all associated skill parameter instances and, if there is no one, returns the None object.

        Returns:
            IndividualValueList: generator with all associated skill parameter instances.
        """
        if len(self.hasParameter) == 0:
            return None
        else:
            return self.hasParameter


class SkillInterface(ExtendedThing):

    # The associated SubmodelElement class of the AAS is also defined
    # aas_sme_class = None
    # aas_sme_class = extended_submodel.ExtendedSkillInterface

    def check_instance(self):
        """
        This method checks whether the SkillInterface instance is valid: if the required attributes are set and if all
        the added properties are valid. In case of invalid SkillInterface, it raises the exception related to the
        checking error.
        """
        pass
        # TODO pensar mas comprobaciones


class SkillParameter(ExtendedThing):

    def is_skill_parameter_type(self, parameter_type_values):
        """
        This method checks whether the SkillParameter instance has one of the given values for the related DataType.

        Args:
            parameter_type_values: the values to check the SkillParameter instance for.

        Returns:
            bool: True if the SkillParameter instance has one of the given values for the related DataType.
        """
        # First, if only one value is passed, a list is created.
        if not isinstance(parameter_type_values, list):
            parameter_type_values = [parameter_type_values]
        for values in self.data_properties_types_dict.values():
            for value in values:
                if value in parameter_type_values:
                    return True
        return False



class StateMachine(ExtendedThing):
    pass


import json
import logging

from spade.behaviour import CyclicBehaviour

from smia.behaviours.specific_handle_behaviours.handle_negotiation_behaviour import HandleNegotiationBehaviour
from smia.behaviours.specific_handle_behaviours.handle_svc_response_behaviour import HandleSvcResponseBehaviour
from smia.css_ontology.css_ontology_utils import CapabilitySkillACLInfo, CapabilitySkillOntologyUtils
from smia.logic import negotiation_utils, inter_aas_interactions_utils
from smia.logic.exceptions import RequestDataError
from smia.utilities import smia_archive_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLJSONSchemas, ServiceTypes
from smia.utilities.smia_info import SMIAInteractionInfo
from smia.utilities.general_utils import GeneralUtils

_logger = logging.getLogger(__name__)


class NegotiatingBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour that handles the negotiation requests made by other standardized SMIAs
    through ACL messages in the I4.0 System.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("NegotiationBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # Wait for a message with the standard ACL template for negotiating to arrive.
        msg = await self.receive(
            timeout=10)  # Timeout set to 10 seconds so as not to continuously execute the behavior.
        if msg:
            # An ACL message has been received by the agent
            _logger.aclinfo("         + Message received on SMIA (NegotiatingBehaviour) from {}".format(msg.sender))
            _logger.aclinfo("                 |___ Message received with content: {}".format(msg.body))

            # The msg body will be parsed to a JSON object
            msg_json_body = json.loads(msg.body)

            # TODO quizas hay que comprobar que el thread no existe? Para hacer al igual que en la peticion de
            #  servicios, utilizar el thread como identificador. Hay que pensarlo, ya que de este modo el thread
            #  serviría para un unico servicio a un AAS

            # Depending on the performative of the message, the agent will have to perform some actions or others
            match msg.get_metadata('performative'):
                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_CFP:
                    _logger.aclinfo("The agent has received a request to start a negotiation (CFP) with thread ["
                                    + msg.thread + "]")

                    # Since the negotiation is an agent capability, it will be checked the validity of the received data
                    cap_req_data = None
                    try:
                        cap_req_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                        await self.check_received_capability_request_data(cap_req_data)
                    except RequestDataError as cap_request_error:
                        # The data are not valid, so the requester must be informed.
                        acl_msg = inter_aas_interactions_utils.create_inter_smia_response_msg(
                            receiver=GeneralUtils.get_sender_from_acl_msg(msg),
                            thread=msg.thread,
                            performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
                            ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION,   # TODO pensar si definir un NegRequest y NegResponse
                            service_id=msg_json_body['serviceID'],
                            service_type=msg_json_body['serviceType'],
                            service_params=json.dumps({'reason': cap_request_error.message})
                        )
                        await self.send(acl_msg)
                        _logger.warning("Capability request with thread [{}] has invalid data, therefore the requester"
                                        " has been informed".format(msg.thread))
                        return  # The run method is terminated to restart checking for new messages.

                    # First, some useful information is obtained from the msg
                    targets_list = cap_req_data['serviceData']['serviceParams']['targets'].split(',')
                    neg_requester_jid = GeneralUtils.get_sender_from_acl_msg(msg)
                    neg_criteria = cap_req_data['serviceData']['serviceParams'][CapabilitySkillACLInfo.REQUIRED_SKILL_NAME]
                    if len(targets_list) == 1:
                        # There is only one target available (therefore, it is the only one, so it is the winner)
                        _logger.info("The SMIA has won the negotiation with thread [" + msg.thread + "]")

                        # As the winner, it will reply to the sender with the result of the negotiation
                        acl_response_msg = negotiation_utils.create_neg_response_msg(
                            receiver=neg_requester_jid,
                            thread=msg.thread,
                            service_id=msg_json_body['serviceID'],
                            service_type=msg_json_body['serviceType'],
                            winner=str(self.myagent.jid))
                        await self.send(acl_response_msg)
                        _logger.aclinfo("ACL response sent for the result of the negotiation request with thread ["
                                        + msg.thread + "]")

                        # The information will be stored in the log
                        execution_info = {'capName': 'Negotiation',
                                          'capType': CapabilitySkillOntologyUtils.AGENT_CAPABILITY_TYPE,
                                          'participants': str(self.myagent.jid), 'criteria': neg_criteria,
                                          'winner': 'True'}
                        smia_archive_utils.save_completed_svc_log_info(
                            GeneralUtils.get_current_timestamp(), GeneralUtils.get_current_timestamp(),
                            inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg), execution_info,
                            ServiceTypes.CSS_RELATED_SERVICE)

                        # Finally, the data is stored in the SMIA
                        # TODO pensar si esto es necesario (ya se almacena todo en el JSON del Archive)
                        neg_data_json = negotiation_utils.create_neg_json_to_store(
                            neg_requester_jid=neg_requester_jid,
                            participants=msg_json_body['serviceData']['serviceParams']['targets'],
                            neg_criteria=neg_criteria,
                            is_winner=True)
                        await self.myagent.save_negotiation_data(thread=msg.thread, neg_data=neg_data_json)

                    else:
                        # If there are more targets, a management behavior is added to the SMIA, which will be
                        # in charge of this particular negotiation. To this end, some information has to be passed to
                        # the behaviour
                        behaviour_info = {
                            'thread': msg.thread,
                            'neg_requester_jid': neg_requester_jid,
                            'targets': msg_json_body['serviceData']['serviceParams']['targets'],
                            'neg_criteria': neg_criteria
                        }
                        # The FIPA-ACL template added to this behavior ensures that you will only receive PROPOSE
                        # messages but also only with the thread of that specific thread
                        handle_neg_template = SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE_PROPOSE
                        handle_neg_template.thread = msg.thread
                        neg_req_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                        handle_neg_behav = HandleNegotiationBehaviour(self.agent, behaviour_info, neg_req_data)
                        self.myagent.add_behaviour(handle_neg_behav, handle_neg_template)

                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
                    _logger.aclinfo("The agent has received an Inter AAS interaction message related to a negotiation:"
                                    " Inform")
                    if msg_json_body['serviceID'] == 'negotiationResult':
                        _logger.aclinfo("The agent has received the result of the negotiation with thread ["+
                                        msg.thread+"].")
                        # TODO BORRAR (enfoque antiguo): pensar como gestionar las respuestas de
                        # As the result of a negotiation is a response to a previous request, a new
                        # HandleSvcResponseBehaviour to handle this service response will be added to the agent
                        svc_resp_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                        svc_resp_handling_behav = HandleSvcResponseBehaviour(self.agent,
                                                                             'Inter AAS interaction',
                                                                             svc_resp_data)
                        self.myagent.add_behaviour(svc_resp_handling_behav)
                    else:
                        _logger.warning('serviceID not available')

                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE:
                    _logger.aclinfo("The agent has received a response in a negotiation: Failure")
                    # TODO pensar como se deberian gestionar este tipo de mensajes en una negociacion
                case _:
                    _logger.error("ACL performative type not available.")
        else:
            _logger.info("         - No message received within 10 seconds on SMIA (NegotiatingBehaviour)")


    async def check_received_capability_request_data(self, cap_req_data):
        """
        This method checks whether the data received contains the necessary information to be able to execute
        the agent capability. If an error occurs, it throws a CapabilityDataError exception.

        Args:
            cap_req_data: (dict): all the information about the agent capability request
        """
        # First, the structure and attributes of the received data are checked and validated
        await inter_aas_interactions_utils.check_received_request_data_structure(
            cap_req_data, ACLJSONSchemas.JSON_SCHEMA_CAPABILITY_REQUEST)
        received_cap_data = cap_req_data['serviceData']['serviceParams']
        cap_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME]
        cap_ontology_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(cap_name)
        if cap_ontology_instance is None:
            raise RequestDataError("The capability {} does not an instance defined in the ontology of this "
                                   "DT".format(cap_name))
        if CapabilitySkillACLInfo.REQUIRED_SKILL_NAME not in received_cap_data:
            raise RequestDataError("The skill need to be defined in negotiation requests.")
        skill_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_NAME]
        result, skill_instance = cap_ontology_instance.check_and_get_related_instance_by_instance_name(skill_name)
        if result is False:
            raise RequestDataError("The capability {} and skill {} are not linked in the ontology of this "
                                   "DT, or the skill does not have an instance"
                                   ".".format(cap_name, skill_name))
        if CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME in received_cap_data:
            # Solo si se ha definido la skill se define la skill interface, sino no tiene significado
            # TODO pensar la frase anterior. Realmente tiene significado o no? Si no se define la skill, podriamos
            #  definir una interfaz que queremos utilizar si o si? En ese caso, habria que buscar una skill con esa
            #  interfaz para ejecutarla
            skill_interface_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME]
            result, instance = skill_instance.check_and_get_related_instance_by_instance_name(skill_interface_name)
            if result is False:
                raise RequestDataError("The skill {} and skill interface {} are not linked in the ontology of "
                                       "this DT, or the skill interface does not have an instance"
                                       ".".format(skill_name, skill_interface_name))
        # Negotiation requests require some mandatory values
        if 'targets' not in received_cap_data:
            raise RequestDataError("The negotiation capability request cannot be fulfilled due to lack of objectives "
                                   "in the negotiation data.")


import logging

from owlready2 import get_ontology, OwlReadyOntologyParsingError, sync_reasoner_pellet, \
    OwlReadyInconsistentOntologyError, ThingClass, Ontology, destroy_entity

from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyUtils
from smia.logic.exceptions import OntologyCheckingAttributeError, OntologyCheckingPropertyError, \
    OntologyInstanceCreationError, OntologyReadingError, CriticalError

_logger = logging.getLogger(__name__)


class CapabilitySkillOntology:
    """
    This class contains related methods of the Capability-Skill ontology.
    """

    def __init__(self):
        self.ontology: Ontology = None

    async def initialize_ontology(self):
        """
        This method initializes the Capability-Skill ontology loading the definition of the OWL ontology in the file
        stored in the config folder.
        """

        # Import only ontologies in RDF/XML, OWL/XML or NTriples format (others do not work, e.g. Turtle)

        self.ontology = get_ontology(CapabilitySkillOntologyUtils.get_ontology_file_path())
        try:
            self.ontology.load()
        except FileNotFoundError as e:
            self.ontology = None
            raise CriticalError("The OWL file of the ontology does not exist.")
        except OwlReadyOntologyParsingError as e:
            if 'NTriples' in e.args[0]:
                raise CriticalError("ERROR: The OWL file is invalid (only RDF/XML, OWL/XML or NTriples format are accepted).")
            if 'Cannot download' in e.args[0]:
                raise CriticalError("ERROR: The OWL has defined imported ontologies that cannot be downloaded.")
            self.ontology = None
            return
        # The ontology has been loaded
        _logger.info("CSS ontology initialized")

    async def execute_ontology_reasoner(self, debug=False):
        """
        This method executes the reasoner for the ontology.

        Args:
            debug (bool): if it is true, the inconsistency of the ontology is explained.
        """
        debug_value = 1
        if debug is True:
            debug_value = 2

        try:
            with self.ontology:
                sync_reasoner_pellet(debug=debug_value)
        except OwlReadyInconsistentOntologyError as e:
            print("ERROR: INCONSISTENT ONTOLOGY!")
            print(e)
            raise OwlReadyInconsistentOntologyError(e)

    async def get_ontology_class_by_iri(self, class_iri):
        """
        This method gets the class within the ontology by its IRI.

        Args:
            class_iri (str): the IRI of the class to be found.

        Returns:
            object: ontology class object.
        """
        result_classes = self.ontology.search(iri=class_iri)
        if len(result_classes) == 0:
            _logger.error("ERROR: class not found with IRI [{}]".format(class_iri))
            return None
        if len(result_classes) > 1:
            _logger.warning("WARNING: THERE IS MORE THAN ONE CLASS WITH IRI [{}], BE CAREFUL".format(class_iri))
        return result_classes[0]

    async def create_ontology_object_instance(self, class_object, instance_name):
        """
        This method creates a new object instance (individual) within the ontology. To this end, a ThingClass is
        required.
        
        Args:
            class_object (ThingClass): ontology class of the instance.
            instance_name (str): name of the instance.

        Returns:
            ThingClass: created instance object.
        """
        if not isinstance(class_object, ThingClass):
            raise OntologyInstanceCreationError("ERROR: the instance cannot be created because the given constructor"
                                                " is not of ThingClass.")
        else:
            # The object of the class is used as constructor of the instance
            return class_object(instance_name)

    async def get_ontology_instance_by_name(self, instance_name):
        """
        This method returns an object instance within the ontology by its name.

        Args:
            instance_name (str): name of the instance.

        Returns:
            ThingClass: class of the instance to be found (None if it is not found).
        """
        for instance_class in self.ontology.individuals():
            if instance_class.name == instance_name:
                return instance_class
        return None

    async def get_ontology_instances_by_class_iri(self, class_iri):
        """
        This method gets all the instances within the ontology with a given class IRI.

        Args:
            class_iri (str): the IRI of the class to be found.

        Returns:
            list: list of ontology instance objects.
        """
        instances_list = []
        for instance in self.ontology.individuals():
            for parent_class in instance.is_a:
                if parent_class.iri == class_iri:
                    instances_list.append(instance)
        if len(instances_list) == 0:
            return None
        return instances_list


    async def add_object_property_to_instances_by_names(self, object_property_name, domain_object_name, range_object_name):
        """
        This method adds a new ObjectProperty to link two instances within the ontology. Checks are performed to ensure
        that the ObjectProperty is valid to link the instances.

        Args:
            object_property_name (str): name of the ObjectProperty to be added.
            domain_object_name (str): name of the instance to which the ObjectProperty will be added.
            range_object_name (str): name of the instance to be added to the domain instance within the given property.
        """
        # Preliminary checks are performed
        if domain_object_name is None or range_object_name is None:
            raise OntologyReadingError("The object property cannot be added because the domain or range are invalid.")
        domain_instance = await self.get_ontology_instance_by_name(domain_object_name)
        range_instance = await self.get_ontology_instance_by_name(range_object_name)
        try:
            getattr(domain_instance, object_property_name).append(range_instance)
        except AttributeError as e:
            raise OntologyReadingError("ERROR: The class {} does not have the attribute {}".format(domain_instance, object_property_name))

    async def check_instance_by_name(self, instance_name):
        """
        This method checks an instance and removes the invalid objects. In case the instance itself is invalid, it is
        removed from the ontology. If any of the properties are invalid they are removed from this instance.

        Args:
            instance_name (str): name of the instance to be checked.
        """
        instance_object = await self.get_ontology_instance_by_name(instance_name)
        if instance_object is None:
            return
        try:
            instance_object.check_instance()
        except OntologyCheckingAttributeError as e:
            destroy_entity(e.invalid_instance)
        except OntologyCheckingPropertyError as e:
            getattr(instance_object, e.concerned_property_name).remove(e.invalid_instance)

    async def get_all_subclasses_iris_of_class(self, owl_class):
        """
        This method gets all IRIs of the subclasses of a given OWL class.

        Args:
            owl_class (ThingClass): OWL class of which the subclasses are to be found.

        Returns:
            list(str): list of IRIs of the subclasses of the given OWL class.
        """
        subclasses = list(owl_class.subclasses())
        all_subclasses_iris = [subclass.iri for subclass in subclasses]
        # The method need to be executed recursively to get all existing subclasses
        for subclass in subclasses:
            all_subclasses_iris.extend(await self.get_all_subclasses_iris_of_class(subclass))
        return all_subclasses_iris



import abc
import logging
import types

import basyx.aas.model
from basyx.aas.model import SubmodelElementList, SubmodelElement, Operation, Submodel, RelationshipElement, \
    AnnotatedRelationshipElement, BasicEventElement, SubmodelElementCollection, Property, MultiLanguageProperty, \
    Range, Blob, File, ReferenceElement, Capability

from smia import AASModelExtensionUtils
from smia.aas_model.extended_aas import ExtendedGeneralMethods
from smia.logic.exceptions import AASModelOntologyError, AASModelReadingError
from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyInfo
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class ExtendedSubmodel(Submodel):
    """This class contains methods to be added to Submodel class of Basyx Python SDK model."""

    def print_submodel_information(self):
        print("Submodel information:")
        print("\tid: {}".format(self.id))
        print("\tid_short: {}".format(self.id_short))
        print("\tdisplayName: {}".format(self.display_name))
        print("\tdescription: {}".format(self.description))
        print("\tcategory: {}".format(self.category))
        print("\tkind: {}".format(self.kind))
        print("\tsemanticId: {}".format(self.semantic_id))
        print("\tsupplementalSemanticId: {}".format(self.supplemental_semantic_id))
        print("\tadministration: {}".format(ExtendedGeneralMethods.print_administration(self.administration)))
        print("\textension: " + "{}".format(ExtendedGeneralMethods.print_namespace_set(self.extension)))
        print("\tdataSpecifications: " + "{}".format(
            ExtendedGeneralMethods.print_data_specifications(self.embedded_data_specifications)))
        print("\tqualifiers: " + "{}".format(ExtendedGeneralMethods.print_namespace_set(self.qualifier)))


class ExtendedSubmodelElement(SubmodelElement):
    """This class contains methods to be added to SubmodelElement class of Basyx Python SDK model."""

    def print_submodel_element_information(self):
        print("Submodel element information:")
        print("\tid_short: {}".format(self.id_short))
        print("\tdescription: {}".format(self.description))
        print("\tsemanticId: {}".format(self.semantic_id))
        print("\tsupplementalSemanticId: {}".format(self.supplemental_semantic_id))
        print("\tqualifiers: " + "{}".format(ExtendedGeneralMethods.print_namespace_set(self.qualifier)))

    def get_qualifier_by_semantic_id(self, semantic_id_reference):
        if self.qualifier is None:
            return None
        else:
            for qualifier in self.qualifier:
                if qualifier.check_semantic_id_exist(semantic_id_reference):
                    return qualifier
            return None

    def get_parent_submodel(self):
        """
        This method gets the submodel where the self SubmodelElement is defined.

        Returns:
            basyx.aas.model.Submodel: parent submodel in form of a Python object.
        """
        if isinstance(self.parent, basyx.aas.model.Submodel):
            return self.parent
        else:
            return self.parent.get_parent_submodel()

    def check_cap_skill_ontology_semantics_and_qualifiers(self):
        """
        This method checks if the SubmodelElement of the Skill has the required semanticIDs and qualifiers defined in
        the Capability-Skill ontology.

        Returns:
            bool: result of the check (only True if both semanticIDs and qualifiers of Capability-Skill ontology exist).
        """
        # It will be checked if the semantic id of the skill is valid within the ontology
        if self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI) is False:
            _logger.error("The skill {} has not valid semanticID regarding the "
                          "Capability-Skill ontology.".format(self))
            return False

        if self.check_cap_skill_ontology_qualifier_for_skills() is False:
            _logger.error("The skill {} has not valid qualifiers regarding the "
                          "Capability-Skill ontology.".format(self))
            return False
        return True

    def check_cap_skill_ontology_qualifier_for_skills(self):
        """
        This method checks if the SubmodelElement of the Skill has valid qualifiers defined in the Capability-Skil
        ontology.

        Returns:
            bool: result of the check (only True if the qualifier of Capability-Skill ontology exists).
        """
        skill_qualifier = self.get_qualifier_by_type('SkillImplementationType')
        if skill_qualifier is not None:
            if skill_qualifier.value in ['STATE', 'TRIGGER', 'OPERATION', 'FUNCTIONBLOCK']:
                return True
        _logger.error("ERROR: the qualifier is not valid in the skill {}".format(self))
        return False

    def get_parent_ref_by_semantic_id(self, semantic_id):
        """
        This method gets the reference of a parent element of the SubmodelElement by the semanticID.

        Args:
            semantic_id (str): semantic identifier of the parent element.

        Returns:
            basyx.aas.model.ModelReference: model reference of the parent element (None if the parent does not exist)
        """
        parent_elem = self.parent
        while parent_elem:
            if parent_elem.check_semantic_id_exist(semantic_id):
                return basyx.aas.model.ModelReference.from_referable(parent_elem)
            else:
                parent_elem = parent_elem.parent
        return None


class ExtendedRelationshipElement(RelationshipElement):

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("\tSpecific attributes of RelationshipElements:")
        print("\t\tfirst: {}".format(self.first))
        print("\t\tsecond: {}".format(self.second))


class ExtendedAnnotatedRelationshipElement(AnnotatedRelationshipElement):

    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("\t\tannotation: {}".format(self.annotation))

    def check_cap_skill_ontology_semantic_id(self):
        """
        This method checks if the Capability has one of the required semanticIDs defined in the Capability-Skill
        ontology, exactly for Capabilities.

        Returns:
            bool: result of the check (only True if the semanticID of Capability-Skill ontology exists).
        """
        if ((self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_IRI))
                or (self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_ASSET_CAPABILITY_IRI))
                or (self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI))):
            return True
        else:
            _logger.error("ERROR: the capability is not valid within the ontology.")
            return False

    def check_cap_skill_ontology_qualifiers(self):
        """
        This method checks if the Capability has valid qualifiers, defined in the Capability-Skil ontology.

        Returns:
            bool: result of the check (only True if the qualifier of Capability-Skill ontology exists).
        """
        capability_qualifier = self.get_qualifier_by_type('ExpressionSemantic')
        if capability_qualifier is not None:
            if capability_qualifier.value in ['REQUIREMENT', 'OFFER', 'ASSURANCE']:
                return True
        _logger.error("ERROR: the qualifier is not valid in the capability {}".format(self))
        return False

    def get_capability_type_in_ontology(self):
        """
        This method gets the type of the capability within the Capability-Skill ontology.

        Returns:
            str: value of the type of the capability within the Capability-Skill ontology.
        """
        if self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_IRI):
            return 'ManufacturingCapability'
        elif self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_ASSET_CAPABILITY_IRI):
            return 'AssetCapability'
        elif self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI):
            return 'AgentCapability'
        else:
            _logger.error("ERROR: the capability type is not valid within the ontology.")
            return None

    def get_semantic_id_of_css_ontology(self):
        """
        This method gets the semanticID of the capability within the Capability-Skill ontology.

        Returns:
            str: value of the semanticID of the capability within the Capability-Skill ontology.
        """
        if self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_IRI):
            return CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_IRI
        elif self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI):
            return CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI
        elif self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_ASSET_CAPABILITY_IRI):
            return CapabilitySkillOntologyInfo.CSS_ONTOLOGY_ASSET_CAPABILITY_IRI
        else:
            raise AASModelOntologyError("The capability {} does not have a valid semanticID within the "
                                        "ontology.".format(self.id_short), self, "OntologySemanticIdMissing")


class ExtendedOperation(Operation):

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("\tSpecific attributes of Operations:")
        print("\t\tinputVariable: {}".format(ExtendedGeneralMethods.print_namespace_set(self.input_variable)))
        print("\t\toutputVariable: {}".format(ExtendedGeneralMethods.print_namespace_set(self.output_variable)))
        print("\t\tinoutputVariable: {}".format(ExtendedGeneralMethods.print_namespace_set(self.in_output_variable)))

    def get_variable_value_id(self, value_id):
        """
        This method gets the variable of the Operation that matches with the given valueId.

        Args:
             value_id (str): the value id of the variable to find.

        Returns:
            (str): id_short of the variable
        """
        all_variables = [self.input_variable, self.output_variable, self.in_output_variable]
        for var_type_set in all_variables:
            for var in var_type_set:
                if var.value_id:
                    for key in var.value_id.key:
                        if key.value == value_id:
                            return var.id_short
        return None

    def get_operation_variables_by_semantic_id(self, semantic_id):
        """
        This method gets all operation variables that have the given semanticID.

        Args:
            semantic_id (str):  semantic identifier of the operation variables to find.

        Returns:
            list: all valid operation variables in form of a list of SubmodelElements.
        """
        operation_variables = []
        all_var_sets = [self.input_variable, self.output_variable, self.in_output_variable]
        for var_set in all_var_sets:
            for operation_variable in var_set:
                if operation_variable.check_semantic_id_exist(semantic_id):
                    operation_variables.append(operation_variable)
        return operation_variables


class ExtendedBasicEventElement(BasicEventElement):

    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("Specific attributes of BasicEventElements:")
        # TODO


class ExtendedEntity:
    # class ExtendedEntity(Entity):

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("\tSpecific attributes of Entities:")
        print("\t\tstatement: {}".format(self.statement))
        print("\t\tentityType: {}".format(self.entity_type))
        print("\t\tglobalAssetId: {}".format(self.global_asset_id))
        print("\t\tspecificAssetId: {}".format(self.specific_asset_id))


class ExtendedSubmodelElementList(SubmodelElementList):

    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("Specific attributes of SubmodelElementLists:")
        # TODO


class ExtendedSubmodelElementCollection(SubmodelElementCollection):

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("\tSpecific attributes of SubmodelElementCollections:")
        print("\t\tvalue: (SME in the collection)")
        for sm_elem_in_collection in self.value:
            print("\t\t\tSubmodelElement: {}".format(sm_elem_in_collection))

    def get_sm_element_by_id_short(self, id_short):
        """
        This method gets a submodel element inside a SubmodelElementCollection by its id_short.
        Args:
            id_short (str): id_short of the submodel element to find.

        Returns:
           basyx.aas.model.SubmodelElement: submodel element in form of Python object.
        """
        return self.value.get('id_short', id_short)

    def get_sm_element_by_semantic_id(self, semantic_id_ref):
        """
        This method gets a submodel element inside a SubmodelElementCollection by its semantic identifier.
        Args:
            semantic_id_ref (str): semantic identifier of the submodel element to find.

        Returns:
           basyx.aas.model.SubmodelElement: submodel element in form of Python object.
        """
        for sm_elem in self.value:
            for reference in sm_elem.semantic_id.key:
                if reference.value == semantic_id_ref:
                    return sm_elem
        return None


# ------------
# DataElements
# ------------
class ExtendedProperty(Property):
    """This class contains methods to be added to Property class of Basyx Python SDK model."""

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("Specific attributes of Properties:")
        print("\t\tvalueType: {}".format(self.value_type))
        print("\t\tvalueId: {}".format(self.value_id))
        print("\t\tvalue: {}".format(self.value))


class ExtendedMultiLanguageProperty(MultiLanguageProperty):

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("Specific attributes of MultiLanguageProperties:")
        print("\t\tvalueId: {}".format(self.value_id))
        print("\t\tvalue: {}".format(self.value))


class ExtendedRange(Range):

    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("Specific attributes of Ranges:")
        # TODO

    def check_value_within_range(self, value):
        """
        This method checks whether a given value is within the range defined in this element.

        Args:
            value: data to be checked.

        Returns:
            bool: the result of the check.
        """
        try:
            formatted_value = value
            if issubclass(self.value_type, basyx.aas.model.datatypes.Float):
                formatted_value = float(value)
            elif issubclass(self.value_type, basyx.aas.model.datatypes.Int):
                formatted_value = int(value)
            # TODO Think about more types
            return True if self.min <= formatted_value <= self.max else False
        except ValueError as e:
            _logger.error("Value error with data {} of type {} in Range {}".format(value, self,value, self))
            return False

class ExtendedBlob(Blob):

    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("Specific attributes of Blobs:")
        # TODO


class ExtendedFile(File):

    def print_submodel_element_information(self):
        ExtendedSubmodelElement.print_submodel_element_information(self)
        print("Specific attributes of Files:")
        print("\t\tvalue: {}".format(self.value))
        print("\t\tcontentType: {}".format(self.content_type))


class ExtendedReferenceElement(ReferenceElement):

    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("Specific attributes of ReferenceElements:")
        # TODO


# CLASSES FOR CSS ONTOLOGY-BASED SOFTWARE
# ----------------------------------------
class ExtendedGenericCSSClass(metaclass=abc.ABCMeta):

    def add_old_sme_class(self, sme_class):
        """
        This method adds the old Basyx SubmodelElement class to be stored to the correct execution of the software. It
        also adds the inheritance to the Extended SMIA class with the method 'add_inheritance_of_extended_class'.

        Args:
            sme_class (basyx.aas.model.SubmodelElement): old submodel element class in BaSyx Python structure.
        """
        self.old_sme_class = sme_class
        # if issubclass(self.old_sme_class, basyx.aas.model.Operation):
        #     print("Antes la Skill era una Operation")
        # if issubclass(self.old_sme_class, basyx.aas.model.SubmodelElement):
        #     print("Antes la Skill era una SubmodelElement")

        # The inheritance to the associated Extended SMIA class is also added
        self.add_inheritance_of_extended_class()

    def add_inheritance_of_extended_class(self):
        """
        This method adds inheritance to SMIA Extended classes from the original BaSyx classes, to add the SMIA methods
        specific to the associated class. To do so, it uses the old_sme_class attribute added during the creation of
        any ExtendedGenericCSSClass.
        """
        associated_extended_class = AASModelExtensionUtils.get_extension_classes_dict()[self.old_sme_class]
        if associated_extended_class != self.__class__:
            new_bases = (self.__class__, associated_extended_class)
            new_class = types.new_class(self.__class__.__name__, new_bases, {})
            self.__class__ = new_class

    def get_semantic_id_of_css_ontology(self):
        """
        This method checks the semanticID of the skill within the Capability-Skill ontology.
        """
        pass


class ExtendedCapability(ExtendedGenericCSSClass, Capability):
    def print_submodel_element_information(self):
        super().print_submodel_element_information()
        print("Specific attributes of Capabilities:")

    def check_cap_skill_ontology_semantics_and_qualifiers(self):
        """
        This method checks if the Capability has the required semanticIDs and qualifiers defined in the Capability-Skill
        ontology, exactly for Capabilities.

        Returns:
            bool: result of the check (only True if both semanticIDs and qualifiers of Capability-Skill ontology exist).
        """
        # It will be checked if the semantic id of the capability is valid within the ontology
        if self.check_cap_skill_ontology_semantic_id() is False:
            _logger.error("The capability {} has not valid semanticID regarding the "
                          "Capability-Skill ontology.".format(self))
            return False

        # It will also be checked if it has any of the qualifiers defined in the ontology for the capabilities
        if self.check_cap_skill_ontology_qualifiers() is False:
            _logger.error("The capability {} has not valid qualifiers regarding the "
                          "Capability-Skill ontology.".format(self))
            return False
        return True


class ExtendedSkill(ExtendedGenericCSSClass):
    # TODO se ha tenido que separar las skills simples (Operation,Event...) de las complejas (Submodel). De
    #  momento estas no se han implementado pero en el White paper de PLattform Industrie se menciona que las Skills
    #  se pueden implementar mediante FunctionBlock (Submodelo)
    pass


class ExtendedSkillInterface(ExtendedGenericCSSClass):
    # TODO se ha tenido que separar las skills interfaces simples (Operation,Event...) de las complejas (SMC). De
    #  momento se tienen ambas opciones ya que las interfaces de servicios de activos son SMC y las de los servicios de
    #  agente son simples

    def get_associated_asset_interface(self):
        """
        This method gets the asset interface AAS SubmodelElement associated with this skill interface.

        Returns:
            basyx.aas.model.SubmodelElement: submodel element of the asset interface.
        """
        pass


class ExtendedCapabilityConstraint(ExtendedGenericCSSClass, ExtendedSubmodelElement):
# class ExtendedCapabilityConstraint(ExtendedGenericCSSClass, ExtendedSubmodelElement, ExtendedProperty):

    def get_semantic_id_of_css_ontology(self):
        """
        This method checks the semanticID of the skill within the Capability-Skill ontology.
        """
        if not self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_CONSTRAINT_IRI):
            raise AASModelOntologyError("The skill {} does not have the valid semanticID within the "
                                        "ontology.".format(self.id_short), self, "OntologySemanticIdMissing")

    def check_constraint(self, constraint_data):
        """
        This method checks the Capability Constraint against the given data to ensure that it is valid.

        Args:
            constraint_data: data to be checked with the Capability Constraint

        Returns:
            bool: the result of the check
        """
        # Initially, the result is set to invalid.
        result = False
        if isinstance(self, ExtendedRange):
            result = self.check_value_within_range(constraint_data)
            if not result:
                _logger.warning("The data ({}) is invalid in the Capability Constraint {} because it is not within "
                                "the defined range values.".format(constraint_data, self))
        # TODO Think about more types of constraints (properties, references...)
        _logger.info('The Capability Constraint {} has been checked with data ({}), resulting in {}'.format(
            self, constraint_data, result))
        return result


class ExtendedSimpleSkill(ExtendedSkill, ExtendedSubmodelElement):
# class ExtendedSimpleSkill(ExtendedSkill, ExtendedSubmodelElement, ExtendedOperation):

    def get_semantic_id_of_css_ontology(self):
        """
        This method checks the semanticID of the skill within the Capability-Skill ontology.
        """
        if not self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI):
            raise AASModelOntologyError("The skill {} does not have the valid semanticID within the "
                                        "ontology.".format(self.id_short), self, "OntologySemanticIdMissing")


class ExtendedComplexSkill(ExtendedSkill, ExtendedSubmodel):
    def get_semantic_id_of_css_ontology(self):
        """
        This method checks the semanticID of the skill within the Capability-Skill ontology.
        """
        if not self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI):
            raise AASModelOntologyError("The skill {} does not have the valid semanticID within the "
                                        "ontology.".format(self.id_short), self, "OntologySemanticIdMissing")


class ExtendedSkillParameter(ExtendedGenericCSSClass):
# class ExtendedSkillParameter(ExtendedGenericCSSClass, ExtendedProperty):
    pass


class ExtendedSimpleSkillInterface(ExtendedSkillInterface, ExtendedSubmodelElement):
# class ExtendedSimpleSkillInterface(ExtendedSkillInterface, ExtendedOperation, ExtendedSubmodelElement):

    def get_semantic_id_of_css_ontology(self):
        """
        This method checks the semanticID of the skill within the Capability-Skill ontology.
        """
        if not self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_INTERFACE_IRI):
            raise AASModelOntologyError("The skill {} does not have the valid semanticID within the "
                                        "ontology.".format(self.id_short), self, "OntologySemanticIdMissing")


class ExtendedComplexSkillInterface(ExtendedSkillInterface, ExtendedSubmodelElementCollection):

    def get_semantic_id_of_css_ontology(self):
        """
        This method checks the semanticID of the skill within the Capability-Skill ontology.
        """
        if not self.check_semantic_id_exist(CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_INTERFACE_IRI):
            raise AASModelOntologyError("The skill {} does not have the valid semanticID within the "
                                        "ontology.".format(self.id_short), self, "OntologySemanticIdMissing")

    def get_associated_asset_interface(self):
        """
        This method gets the asset interface AAS SubmodelElement associated with this skill interface.

        Returns:
            basyx.aas.model.SubmodelElement: submodel element of the asset interface.
        """
        #
        asset_interface_elem = self.get_parent_ref_by_semantic_id(AssetInterfacesInfo.SEMANTICID_INTERFACE)
        if asset_interface_elem is None:
            raise AASModelReadingError("The skill interface is not inside the AssetInterfaceSubmodel.", self,
                                       "SubmodelElementNotInValidSubmodel")
        return asset_interface_elem






import asyncio
import logging
import basyx.aas.model.submodel
from basyx.aas.util import traversal
from smia.aas_model.aas_model_utils import AASModelInfo

from smia.logic.exceptions import CapabilityCheckingError, AASModelReadingError, OntologyReadingError
from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyUtils, CapabilitySkillACLInfo, \
    CapabilitySkillOntologyInfo
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class ExtendedAASModel:
    """This class contains methods related to the AAS model stored in Python objects. These methods are useful when
    using the AAS model in the SMIA approach."""

    aas_model_object_store = None  #: Storage with all Python object obtained from the AAS model
    capabilities_skills_dict = None  #: Dictionary with all information related to Capability-Skill model obtained from the AAS model

    lock = None  #: Asyncio Lock object for secure access to shared AAS model objects

    def __init__(self):

        # The object to store all Python objects obtained from the AAS model are initialized
        self.aas_model_object_store = None

        # Also, the object to store all information related to Capabilities and Skills is initialized. To this end, the
        # dictionary is divided in capabilities related to the agent (DT) and capabilities related to the asset.
        self.capabilities_skills_dict = {'AgentCapabilities': {}, 'AssetCapabilities': {}}

        # The Lock object is used to manage the access to global agent attributes (request and response dictionaries,
        # interaction id number...)
        self.lock = asyncio.Lock()

    # -----------------------------------------
    # Methods related to AAS model object store
    # -----------------------------------------
    async def set_aas_model_object_store(self, object_store):
        """
        This method updates the object store for the AAS model.

        Args:
            object_store (basyx.aas.model.DictObjectStore): object to store all Python elements of the AAS model.

        """
        async with self.lock:
            self.aas_model_object_store = object_store

    async def get_aas_model_object_store(self):
        """
        This method returns the object store for the AAS model.

        Returns:
            basyx.aas.model.DictObjectStore: object with all Python elements of the AAS model.

        """
        async with self.lock:
            return self.aas_model_object_store

    # ---------------
    # General methods
    # ---------------
    async def execute_general_analysis(self, smia_agent):
        """
        This method performs the general analysis on the AAS model to check if it is a valid AAS model for SMIA
        software, and provides information to the user via console or log file.

        Args:
            smia_agent(spade.agent.Agent): SMIA SPADE agent object.
        """
        # If standardized submodels are added is checked
        asset_interfaces_submodel = await self.get_submodel_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL)
        if not asset_interfaces_submodel:
            _logger.warning("AssetInterfacesSubmodel submodel is not defined. Make sure that this SMIA does not need to "
                            "be connected to the asset via any communication protocols.")
        software_nameplate_submodel = await self.get_submodel_by_semantic_id(
            AASModelInfo.SEMANTICID_SOFTWARE_NAMEPLATE_SUBMODEL)
        if not software_nameplate_submodel:
            _logger.warning("SoftwareNameplate submodel is not defined. This can lead to errors during SMIA runtime.")
        else:
            # Checks if the instance name defined in the AAS matches the one defined in the agent during start-up.
            for software_nameplate_instance in software_nameplate_submodel.submodel_element:
                aas_smia_instance_name = software_nameplate_instance.get_sm_element_by_semantic_id(AASModelInfo.SEMANTICID_SOFTWARE_NAMEPLATE_INSTANCE_NAME)
                if aas_smia_instance_name is not None:
                    if aas_smia_instance_name.value != smia_agent.jid.localpart:
                        _logger.warning(
                            "The SMIA instance name defined in the SoftwareNameplate submodel [{}] and the identifier "
                            "used to start the agent [{}] do not match. This can result in errors during SMIA runtime."
                            "".format(aas_smia_instance_name.value, smia_agent.jid.localpart))
        # TODO Add more checks


    # -------------------------------------------
    # Methods related to capability skills object
    # -------------------------------------------
    async def set_capabilities_skills_object(self, cap_skills_object):
        """
        This method updates the object that contains with all the information about Capabilities and Skills of the DT
        and the asset.

        Args:
            cap_skills_object (dict): object with all the information about Capabilities and Skills of the DT and the asset.

        """
        async with self.lock:
            self.capabilities_skills_dict = cap_skills_object

    async def get_capabilities_skills_object(self):
        """
        This method returns the object with all the information about Capabilities and Skills of the DT and the asset.

        Returns:
            dict: all the information about Capabilities and Skills of the DT and the asset in JSON format.

        """
        async with self.lock:
            return self.capabilities_skills_dict

    async def save_capability_skill_information(self, capability_type, cap_skill_info):
        """
        This method saves the information of a capability and its associated skill in the global dictionary. It
        distinguishes between AgentCapabilities and AssetCapabilities.

        Args:
            capability_type (str): type of the capability (AgentCapabilities or AssetCapabilities).
            cap_skill_info (dict): information in form of a JSON object.
        """
        if capability_type not in CapabilitySkillOntologyUtils.CAPABILITY_TYPE_POSSIBLE_VALUES:
            _logger.error("The capability type is not valid. The information cannot be saved.")
            return
        async with self.lock:
            try:
                if capability_type == CapabilitySkillOntologyUtils.AGENT_CAPABILITY_TYPE:
                    self.capabilities_skills_dict['AgentCapabilities'] = cap_skill_info
                if capability_type == CapabilitySkillOntologyUtils.ASSET_CAPABILITY_TYPE:
                    self.capabilities_skills_dict['AssetCapabilities'] = cap_skill_info
            except KeyError:
                _logger.error("The capability type is not valid. The information cannot be saved.")

    async def get_capability_dict_by_type(self, cap_type):
        """
        This method returns the capability dictionary related to the given capability type.

        Args:
            cap_type(str): type of the capability (AgentCapabilities or AssetCapabilities).

        Returns:
            dict: dictionary will the information of all capabilities of the given type.
        """
        async with self.lock:
            if cap_type == CapabilitySkillOntologyUtils.AGENT_CAPABILITY_TYPE:
                return self.capabilities_skills_dict['AgentCapabilities']
            if cap_type == CapabilitySkillOntologyUtils.ASSET_CAPABILITY_TYPE:
                return self.capabilities_skills_dict['AssetCapabilities']
            return {}

    # -----------------------------------------
    # General methods related to AAS meta-model
    # -----------------------------------------
    async def get_object_by_reference(self, reference):
        """
        This method gets the AAS meta-model Python object using the reference, distinguishing between ExternalReference
         and ModelReference.

        Args:
            reference (basyx.aas.model.Reference): reference object related to desired element

        Returns:
            object: Python object of the desired element associated to the reference.
        """
        try:
            if isinstance(reference, basyx.aas.model.ExternalReference):
                for key in reference.key:
                    return self.aas_model_object_store.get_identifiable(key.value)
            elif isinstance(reference, basyx.aas.model.ModelReference):
                return reference.resolve(self.aas_model_object_store)
        except KeyError as e:
            _logger.error(e)
            raise AASModelReadingError("The object within the AAS model with reference {} does not "
                                       "exist".format(reference), sme_class=None, reason='AASModelObjectNotExist')

    async def get_submodel_elements_by_semantic_id(self, semantic_id_external_ref, sme_class=None):
        """
        This method gets all SubmodelElements by the semantic id in form of an external reference. The SubmodelElements
        to obtain can be filtered by the meta-model class.

        Args:
            semantic_id_external_ref (str): semantic id in form of an external reference
            sme_class (basyx.aas.model.SubmodelElement): Submodel Element class of the elements to be found (None if no
             filtering is required).

        Returns:
            list(basyx.aas.model.SubmodelElement): list with all SubmodelElements of the given class.
        """
        if sme_class is None:
            sme_class = basyx.aas.model.SubmodelElement
        rels_elements = []
        for aas_object in self.aas_model_object_store:
            if isinstance(aas_object, basyx.aas.model.Submodel):
                for submodel_element in traversal.walk_submodel(aas_object):
                    if isinstance(submodel_element, sme_class):
                        if submodel_element.check_semantic_id_exist(semantic_id_external_ref):
                            rels_elements.append(submodel_element)
                        if isinstance(submodel_element, basyx.aas.model.Operation):
                            # In case of Operation, OperationVariables need to be analyzed
                            rels_elements.extend(submodel_element.get_operation_variables_by_semantic_id(
                                semantic_id_external_ref))
        return rels_elements

    async def get_submodel_elements_by_semantic_id_list(self, semantic_id_external_refs, sme_class=None):
        """
        This method obtains all the SubmodelElements that have any of the given semantic identifiers (in form of an
        external references). The SubmodelElements to obtain can be filtered by the meta-model class.

        Args:
            semantic_id_external_refs (list(str)): semantic identifiers in form of a list of external references
            sme_class (basyx.aas.model.SubmodelElement): Submodel Element class of the elements to be found (None if no
             filtering is required).

        Returns:
            list(basyx.aas.model.SubmodelElement): list with all SubmodelElements of the given class.
        """
        if sme_class is None:
            sme_class = basyx.aas.model.SubmodelElement
        rels_elements = []
        for aas_object in self.aas_model_object_store:
            if isinstance(aas_object, basyx.aas.model.Submodel):
                for submodel_element in traversal.walk_submodel(aas_object):
                    if isinstance(submodel_element, sme_class):
                        for semantic_id in semantic_id_external_refs:
                            if submodel_element.check_semantic_id_exist(semantic_id):
                                rels_elements.append(submodel_element)
                            if isinstance(submodel_element, basyx.aas.model.Operation):
                                # In case of Operation, OperationVariables need to be analyzed
                                rels_elements.extend(submodel_element.get_operation_variables_by_semantic_id(
                                    semantic_id))
        return rels_elements

    async def get_submodel_by_semantic_id(self, sm_semantic_id):
        """
        This method gets the Submodel object using its semantic identifier.

        Args:
            sm_semantic_id (str): semantic identifier of the Submodel.

        Returns:
            basyx.aas.model.Submodel: Submodel in form of a Python object.
        """
        for aas_object in self.aas_model_object_store:
            if isinstance(aas_object, basyx.aas.model.Submodel) and aas_object.semantic_id is not None:
                for reference in aas_object.semantic_id.key:
                    if reference.value == sm_semantic_id:
                        return aas_object

    async def check_element_exist_in_namespaceset_by_id_short(self, namespaceset_elem, elem_id_short):
        """
        This method checks if an element exists in the NamespaceSet using its id_short.

        Args:
            namespaceset_elem (basyx.aas.model.NamespaceSet): NamespaceSet element
            elem_id_short (str): id_short of the element.

        Returns:
            bool: result of the check
        """
        for namespace_elem in namespaceset_elem:
            if namespace_elem.id_short == elem_id_short:
                return True
        return False

    async def get_concept_description_pair_value_id_by_value_name(self, concept_description_id, value_name):
        """
        This method gets the value_id of a pair within a Concept Description using the value name.

        Args:
            concept_description_id (str): globally unique identifier of the Concept Description.
            value_name (str): name of the value inside the pair to find.

        Returns:
            str: value_id of the pair that contains the provided value name.
        """
        concept_description = self.aas_model_object_store.get_identifiable(concept_description_id)
        if concept_description:
            # First, it is checked the embedded_data_specifications (the specification of the data inside the element).
            if concept_description.embedded_data_specifications:
                # Vamos a comprobar que tenga valueList (en esa variable se añaden los posibles valores para una propiedad)
                for embedded_data_spec in concept_description.embedded_data_specifications:
                    if isinstance(embedded_data_spec.data_specification_content,
                                  basyx.aas.model.DataSpecificationIEC61360):
                        if embedded_data_spec.data_specification_content.value_list:
                            value_list = embedded_data_spec.data_specification_content.value_list
                            for value_elem in value_list:
                                if value_elem.value == value_name:
                                    # As it is another ConceptDescription, it is a ExternalReference (first key)
                                    return value_elem.value_id.key[0].value

        _logger.error("Concept Description with id [{}] not found.".format(concept_description_id))
        return None

    # ---------------------------------------------------------------
    # Methods related to Capability-Skill ontology and AAS meta-model
    # ---------------------------------------------------------------
    async def get_capability_by_id_short(self, cap_type, cap_id_short):
        """
        This method gets the capability object with all its information using its id_short attribute and the type of
        the Capability.

        Args:
            cap_type (str): type of the capability (AgentCapabilities or AssetCapabilities).
            cap_id_short (str): id_short of the Capability to find.

        Returns:
            basyx.aas.model.Capability: Python object of capability to find (None if the Capability does not exist)
        """
        for cap_elem, cap_info in (await self.get_capability_dict_by_type(cap_type)).items():
            if cap_elem.id_short == cap_id_short:
                return cap_elem
        return None

    async def get_cap_skill_elem_from_relationship(self, rel_element):
        """
        This method returns the Capability and Skill objects from the Relationship element, no matter in which order
        they are specified.

        Args:
            rel_element (basyx.aas.model.RelationshipElement): Python object of the RelationshipElement.

        Returns:
            basyx.aas.model.Capability, basyx.aas.model.SubmodelElement: capability and skill SME in Python
            reference objects.
        """
        first_rel_elem = await self.get_object_by_reference(rel_element.first)
        second_rel_elem = await self.get_object_by_reference(rel_element.second)
        if isinstance(first_rel_elem, basyx.aas.model.Capability):
            return first_rel_elem, second_rel_elem
        elif isinstance(second_rel_elem, basyx.aas.model.Capability):
            return second_rel_elem, first_rel_elem
        else:
            _logger.error(
                "This method has been used incorrectly. This Relationship does not have a Capability element.")
            return None, None

    async def get_elements_from_relationship(self, rel_element, first_elem_class=None, second_elem_class=None):
        """
        This method returns the objects of a given Relationship element taking into account the type of class that is
        required for the objects referenced within the relationship. The objects will be returned in the order
        specified by the classes, no matter in which order they are defined in the AAS model (in the case of not
        specifying any class, it is returned in the original order).

        Args:
            rel_element (basyx.aas.model.RelationshipElement): Python object of the RelationshipElement.
            first_elem_class (basyx.aas.model.SubmodelElement): Class required for the first element returned.
            second_elem_class (basyx.aas.model.SubmodelElement): Class required for the second element returned.

        Returns:
            basyx.aas.model.SubmodelElement, basyx.aas.model.SubmodelElement: SME Python objects with the required format.
        """
        # Using the references within the relationship, both SubmodelElement are obtained
        first_rel_elem = await self.get_object_by_reference(rel_element.first)
        second_rel_elem = await self.get_object_by_reference(rel_element.second)
        if first_rel_elem is None or second_rel_elem is None:
            raise AASModelReadingError("Elements of the relationship {} does not exist in the AAS model"
                                       "".format(rel_element.id_short), sme_class=rel_element,
                                       reason="Relationship referenced element invalid")
        if None in (first_rel_elem, second_rel_elem):
            # If no one is required, it simply returns the elements
            return first_rel_elem, second_rel_elem
        if None not in (first_rel_elem, second_rel_elem):
            # If both are required, both have to be checked
            if (first_elem_class is None) or (second_elem_class is None):
                raise OntologyReadingError('The classes of the relation {} do not exist. Check the OWL ontology '
                                           'file.'.format(rel_element.id_short))
            if isinstance(first_rel_elem, first_elem_class) and isinstance(second_rel_elem, second_elem_class):
                return first_rel_elem, second_rel_elem
            elif isinstance(first_rel_elem, second_elem_class) and isinstance(second_rel_elem, first_elem_class):
                return second_rel_elem, first_rel_elem
            else:
                raise AASModelReadingError("Elements of the relationship {} are not exist of the required classes {}, "
                                           "{}".format(rel_element.id_short, first_elem_class, second_elem_class)
                                           , sme_class=rel_element, reason="Relationship referenced element invalid")
        if first_elem_class is not None:
            if isinstance(first_rel_elem, first_elem_class):
                return first_rel_elem, second_rel_elem
            elif isinstance(second_rel_elem, first_elem_class):
                return second_rel_elem, first_rel_elem
            else:
                raise AASModelReadingError("The element {} within the relationship {} is not of the required class "
                                           "{}".format(first_rel_elem, rel_element.id_short, first_elem_class),
                                           sme_class=rel_element, reason="Relationship referenced element invalid")
        if second_elem_class is not None:
            if isinstance(first_rel_elem, second_elem_class):
                return second_rel_elem, first_rel_elem
            elif isinstance(second_rel_elem, second_elem_class):
                return first_rel_elem, second_rel_elem
            else:
                raise AASModelReadingError("The element {} within the relationship {} is not of the required class "
                                           "{}".format(second_rel_elem, rel_element.id_short, second_elem_class),
                                           sme_class=rel_element, reason="Relationship referenced element invalid")

    async def get_capability_associated_constraints(self, capability_elem):
        """
        This method gets the constraints associated to a capability.

        Args:
            capability_elem (basyx.aas.model.Capability): capability Python object.

        Returns:
            list: list with all constraints of the selected capability in form of Python objects.
        """
        cap_constraints = []
        rels_cap_constraints = await self.get_submodel_elements_by_semantic_id(
            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_PROP_ISRESTRICTEDBY_IRI, basyx.aas.model.RelationshipElement)
        for rel in rels_cap_constraints:
            first_elem = await self.get_object_by_reference(rel.first)
            second_elem = await self.get_object_by_reference(rel.second)
            if first_elem == capability_elem:
                cap_constraints.append(second_elem)
            elif second_elem == capability_elem:
                cap_constraints.append(first_elem)
        return cap_constraints

    async def get_capability_associated_constraints_by_qualifier_data(self, capability_elem, qualifier_type,
                                                                      qualifier_value):
        """
        This method gets the constraints associated to a capability that have specific qualifier data.

        Args:
            capability_elem (basyx.aas.model.Capability): capability Python object.
            qualifier_type (str): type of the qualifier
            qualifier_value (str): value of the qualifier

        Returns:
            list: list with all constraints of the selected capability in form of Python objects.
        """
        all_constraints = await self.get_capability_associated_constraints(capability_elem)
        for constraint in all_constraints:
            if constraint.qualifier:
                for qualifier in constraint.qualifier:
                    if (qualifier.type == qualifier_type) and (qualifier.value == qualifier_value):
                        return constraint
        return None

    async def get_skill_interface_by_skill_elem(self, skill_elem):
        """
        This method gets the interfaces associated to a skill.

        Args:
            skill_elem (basyx.aas.model.SubmodelElement): skill Python object in form of a SubmodelElement.

        Returns:
            (basyx.aas.model.SubmodelElement): the interface of the selected skill in form of Python object (None if it does not exist).
        """
        rels_skill_interfaces = await self.get_submodel_elements_by_semantic_id(
            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_IRI, basyx.aas.model.RelationshipElement)
        for rel in rels_skill_interfaces:
            first_elem = await self.get_object_by_reference(rel.first)
            second_elem = await self.get_object_by_reference(rel.second)
            if first_elem == skill_elem:
                return second_elem
            elif second_elem == skill_elem:
                return first_elem
        return None

    async def get_asset_interface_interaction_metadata_by_value_semantic_id(self, value_semantic_id):
        """
        This method reads the AssetInterfacesDescription submodel and returns an Interaction Metadata by a given value
        semanticID. This is how in this approach it is established that an attribute is of asset data type.

        Args:
            value_semantic_id (str): semanticID of the value of the Interaction Metadata.

        Returns:
            basyx.aas.model.SubmodelElementCollection: SubmodelElement of the required Interaction Metadata (None if the semanticID does not exist)
        """
        asset_interfaces_submodel = await self.get_submodel_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL)
        for interface_smc in asset_interfaces_submodel.submodel_element:
            interaction_metadata = interface_smc.get_sm_element_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERACTION_METADATA)
            for element_type_smc in interaction_metadata:
                # InteractionMetadata has properties, actions or events
                for element_smc in element_type_smc:
                    # The valueSemantic SubmodelElement is obtained
                    value_semantics = element_smc.get_sm_element_by_semantic_id(
                        AssetInterfacesInfo.SEMANTICID_VALUE_SEMANTICS)
                    if value_semantics:
                        for reference in value_semantics.value.key:
                            if reference.value == value_semantic_id:
                                return element_smc
        return None

    async def get_skill_parameters_exposure_interface_elem(self, skill_elem):
        """
        This method gets the exposure element within the skill interface linked to the parameters of the given skill.

        Args:
            skill_elem (basyx.aas.model.SubmodelElement): skill Python object in form of a SubmodelElement.

        Returns:
            basyx.aas.model.SubmodelElement: exposure submodel element of skill parameters.
        """
        # The exposure elements can be obtained with the related relationship semanticID
        rels_params_exposed = await self.get_submodel_elements_by_semantic_id(
            CapabilitySkillOntologyUtils.SEMANTICID_REL_SKILL_PARAMETER_SKILL_INTERFACE, basyx.aas.model.RelationshipElement)
        for rel in rels_params_exposed:
            first_elem = await self.get_object_by_reference(rel.first)
            second_elem = await self.get_object_by_reference(rel.second)
            if first_elem == skill_elem:
                 return  second_elem
            elif second_elem == skill_elem:
                return first_elem
        return None

    async def skill_feasibility_checking_post_conditions(self, capability_elem, constraints_data):
        """
        This method checks the feasibility of a Capability element in relation with its post-conditions.

        Args:
            capability_elem (basyx.aas.model.Capability): capability Python object.
            constraints_data (dict): JSON object with the data of the constraints (with required values)
        """
        # First, the postcondition constraints are obtained
        post_condition_constraints = await self.get_capability_associated_constraints_by_qualifier_data(capability_elem,
                                                                                                        CapabilitySkillOntologyUtils.QUALIFIER_FEASIBILITY_CHECKING_TYPE,
                                                                                                        'POSTCONDITION')
        if post_condition_constraints:
            # TODO habra que pensar como analizar las post condiciones
            pass



import logging

from smia.behaviours.check_physical_asset_behaviour import CheckPhysicalAssetBehaviour
from smia.states.state_booting import StateBooting
from smia.utilities.general_utils import SMIAGeneralInfo

_logger = logging.getLogger(__name__)


class StateBootingResource(StateBooting):
    """
    This class contains the Boot state of the common SMIA.
    """

    async def run(self):
        """
        This method implements the boot state of the SMIA of type resource. Here all the required initialization
        tasks are performed.
        """

        # First the common initialization tasks perfomed by any SMIA will be executed
        await super().booting_state_logic()

        # In the booting state of type resource the physical asset has to be checked
        check_physical_asset_behav = CheckPhysicalAssetBehaviour(self.agent)
        self.agent.add_behaviour(check_physical_asset_behav)

        # Wait until the behaviour has finished to pass to running state.
        await check_physical_asset_behav.join()

        _logger.info(f"{self.agent.jid} agent has finished it Boot state.")

        self.set_next_state(SMIAGeneralInfo.RUNNING_STATE_NAME)




"""This class groups the methods related to the Inter AAS interactions between I4.0 SMIA entities."""
import json
import logging

import jsonschema
from jsonschema.exceptions import ValidationError
from spade.message import Message

from smia.logic.exceptions import RequestDataError
from smia.utilities.smia_info import SMIAInteractionInfo
from smia.utilities.general_utils import GeneralUtils

_logger = logging.getLogger(__name__)


def create_svc_json_data_from_acl_msg(acl_msg):
    """
    This method creates the dictionary with all the required data of a service related to an ACL message.

    Args:
        acl_msg (spade.message.Message): ACL message where to get the information

    Returns:
        dict: dictionary with all the information about the service
    """
    svc_req_data_json = {
        'performative': acl_msg.get_metadata('performative'),
        'ontology': acl_msg.get_metadata('ontology'),
        'thread': acl_msg.thread,
        'sender': GeneralUtils.get_sender_from_acl_msg(acl_msg),
        'receiver': str(acl_msg.to),
    }
    # The body of the ACL message contains the rest of the information
    svc_req_data_json.update(json.loads(acl_msg.body))
    return svc_req_data_json

def create_inter_smia_response_msg(receiver, thread, performative, ontology, service_id=None, service_type=None,
                                   service_params=None):
    """
    This method creates the Inter AAS interaction response object.

    Args:
        receiver (str): the JID of the receiver of the ACL message from which the service is requested.
        thread (str): the thread of the ACL message.
        performative (str): the performative of the ACL message.
        ontology (str): the ontology of the ACL message.
        service_id (str): the serviceID of the ACL message.
        service_type (str): the serviceType of the ACL message.
        service_params (str): the serviceParams of the "serviceData" section of the ACL message.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-compliant.
    """

    request_msg = Message(to=receiver, thread=thread)
    request_msg.set_metadata('performative', performative)
    request_msg.set_metadata('ontology', ontology)
    # request_msg.set_metadata('ontology', 'SvcResponse')

    request_msg_body_json = {
        'serviceID': service_id,
        'serviceType': service_type,
        'serviceData': {
            'serviceCategory': 'service-response',
            'timestamp': GeneralUtils.get_current_timestamp(),
        }
    }
    if service_params is not None:
        request_msg_body_json['serviceData']['serviceParams'] = service_params
    request_msg.body = json.dumps(request_msg_body_json)
    return request_msg

async def check_received_request_data_structure(received_data, json_schema):
    """
    This method checks if the received data for a request is valid. The JSON object with the specific
    data is also validated against the given associated JSON Schema. In any case, if it is invalid, it raises a
    RequestDataError exception.

    Args:
        received_data (dict): received data in form of a JSON object.
        json_schema (dict): JSON Schema in form of a JSON object.
    """
    # TODO modificarlo cuando se piense la estructura del lenguaje I4.0
    if 'serviceData' not in received_data:
        raise RequestDataError("The received request is invalid due to missing #serviceData field in the"
                               "request message.")
    if 'serviceParams' not in received_data['serviceData']:
        raise RequestDataError("The received request is invalid due to missing #serviceParams field within "
                               "the #serviceData section of the request message.")
    # The received JSON object is also validated against the associated JSON Schema
    try:
        jsonschema.validate(instance=received_data['serviceData']['serviceParams'],
                            schema=json_schema)
    except ValidationError as e:
        raise RequestDataError("The received JSON data within the request message is invalid against the required "
                               "JSON schema. Invalid part: {}. Reason: {}.".format(e.instance, e.message))


import logging
from spade.behaviour import State

from smia.behaviours.init_aas_model_behaviour import InitAASModelBehaviour
from smia.utilities import smia_archive_utils
from smia.utilities.general_utils import SMIAGeneralInfo

_logger = logging.getLogger(__name__)


class StateBooting(State):
    """
    This class contains the Boot state of the common SMIA.
    """

    async def run(self):
        """
        This method implements the boot state of the common SMIA. Here all the required initialization tasks
        are performed.
        """

        await self.booting_state_logic()
        self.set_next_state(SMIAGeneralInfo.RUNNING_STATE_NAME)

    async def booting_state_logic(self):
        """
        This method contains the logic of the boot state of the common SMIA. This method can be used by any
        inherited class.
        """
        _logger.info("## STATE 1: BOOTING ##  (Initial state)")

        # First, it is ensured that the attributes of the SMIA are initialized
        self.agent.initialize_smia_attributes()

        # The ontology has to be initialized in order to be available during the AAS model analysis
        await self.agent.css_ontology.initialize_ontology()

        # The submodels also have to be initialized, so its behaviour is also added
        init_aas_model_behav = InitAASModelBehaviour(self.agent)
        self.agent.add_behaviour(init_aas_model_behav)

        # Wait until the behaviours have finished because the AAS Archive has to be initialized to pass to running state
        await init_aas_model_behav.join()

        # If the initialization behaviour has completed, SMIA is in the InitializationReady status
        smia_archive_utils.update_status('InitializationReady')

        # Finished the Boot State the agent can move to the next state
        _logger.info(f"{self.agent.jid} agent has finished it Boot state.")



"""This class contains utility methods related to the ConfigMap file."""
import logging
import configparser
import os.path

from smia.logic.exceptions import CriticalError
from smia.utilities.smia_general_info import SMIAGeneralInfo

_logger = logging.getLogger(__name__)

# ---------------
# General methods
# ---------------
def create_empty_file():
    """
    This method creates a properties configuration file for SMIA with default values ('#').
    """

    config_prop = configparser.RawConfigParser()
    config_prop['DT'] = {'dt.version': '#', 'dt.agentID': '#', 'dt.password': '#', 'dt.xmpp-server': '#',
                         'dt.web-ui': '#'}
    config_prop['AAS'] = {'aas.meta-model.version': '#', 'aas.model.serialization': '#', 'aas.model.folder': '#',
                          'aas.model.file': '#'}
    config_prop['ONTOLOGY'] = {'ontology.file': '#', 'ontology.inside-aasx': '#'}
    update_properties_file_by_parser(config_prop)


def update_properties_file_by_parser(new_config_parser):
    """
    This method updates the content of the properties file by a given parser.

    Args:
        new_config_parser (configparser.RawConfigParser): the parser with the new content for the properties file.
    """
    with (open(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME, 'w')
          as propertiesFile):
        new_config_parser.write(propertiesFile)

def update_properties_file_by_bytes(new_config_bytes):
    """
    This method updates the content of the properties file by given properties in the form of bytes.

    Args:
        new_config_bytes (configparser.RawConfigParser): the parser with the new content for the properties file.
    """
    config_parser = configparser.RawConfigParser()
    new_config_str = new_config_bytes.decode('utf-8')
    config_parser.read_string(new_config_str)

    # Only not defined data need to be updated
    archive_config_parser = configparser.RawConfigParser()
    archive_config_parser.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)
    for section in archive_config_parser.sections():
        for key, value in archive_config_parser.items(section):
            if value == '#':
                archive_config_parser[section][key] = config_parser[section][key]

    update_properties_file_by_parser(archive_config_parser)

# ----------------------------------
# Methods related to AAS information
# ----------------------------------
def get_aas_general_property(property_name):
    """
    This method returns the property of the AAS set in the ConfigMap by the AAS Controller during the deployment
    process. This information is stored in "general.properties" file within "AAS" section (with the 'aas.' prefix).

    Args:
        property_name (str): The name of the property.
    Returns:
        str: The general property of the AAS.
    """
    # Read submodels configuration
    config_sm = configparser.RawConfigParser()
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)

    try:
        return config_sm['AAS']['aas.' + property_name]
    except KeyError as e:
        _logger.error("The 'general.properties' file in the ConfigMap is not valid.")
        return None

def set_aas_general_property(property_name, property_value):
    """
    This method sets a new value for a property of the AAS. The information is stored in "general.properties" file
    within "AAS" section (with the 'aas.' prefix).

    Args:
        property_name (str): The name of the property.
        property_value (str): The new value of the property.
    """
    # Read submodels configuration
    config_prop = configparser.RawConfigParser()
    config_prop.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)

    try:
        config_prop['AAS']['aas.' + property_name] = property_value
        update_properties_file_by_parser(config_prop)
    except KeyError as e:
        _logger.error("The 'general.properties' file in the ConfigMap is not valid regarding 'AAS' section.")
        return None

def get_aas_model_filepath():
    """
    This method returns the AAS model file path. The AAS model is specified in the ‘general.properties’ file
    within the ‘AAS’ section, with 'aas.model.file' attribute.

    Returns:
        str: The AAS model file path within the SMIA Archive.
    """
    # config_sm = configparser.RawConfigParser()
    # config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)
    # return SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + config_sm['AAS']['aas.model.file']
    return SMIAGeneralInfo.CONFIGURATION_AAS_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_AAS_MODEL_FILENAME


# --------------------------------------
# Methods related to DT information
# --------------------------------------
def get_dt_general_property(property_name):
    """
    This method returns the DT property set in the ConfigMap during the deployment process. This information is stored
    in the ‘general.properties’ file within the ‘DT’ section (with the 'dt.' prefix).

    Returns:
        str: The general property of the DT.
    """
    config_sm = configparser.RawConfigParser()
    if not os.path.exists(
            SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME):
        raise CriticalError("General properties file not found in [" + SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' +
                            SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME + "], the path is invalid.")
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)
    try:
        return config_sm['DT']['dt.' + property_name]
    except KeyError as e:
        raise CriticalError("The 'general.properties' file in the ConfigMap does not have valid keys.")

# --------------------------------------
# Methods related to ontology information
# --------------------------------------
def get_ontology_general_property(property_name):
    """
    This method returns the property of the OWL set in the ConfigMap by the AAS Controller during the deployment
    process. This information is stored in "general.properties" file within "ONTOLOGY" section (with the 'ontology.'
    prefix).

    Args:
        property_name (str): The name of the property.
    Returns:
        str: The general property of the ontology.
    """
    # Read submodels configuration
    config_sm = configparser.RawConfigParser()
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)

    try:
        return config_sm['ONTOLOGY']['ontology.' + property_name]
    except KeyError as e:
        _logger.error("The 'initialization.properties' file in the SMIA Archive is not valid or it is not defined.")
        return None

def get_defined_ontology_filepath():
    """
    This method returns the OWL ontology file path. If the file for the OWL ontology is specified in the
    ‘general.properties’ file, it is located within the ‘ONTOLOGY’ section, with 'ontology.owl.file' attribute.

    Returns:
        str: The ontology file path within the SMIA Archive.
    """
    config_sm = configparser.RawConfigParser()
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME)
    return SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + config_sm['ONTOLOGY']['ontology.file']

def create_ontology_file(file_bytes_content):
    """
    This method creates the ontology OWL file with the given file content. This method is used when the ontology file
    is inside the AASX package, so it must be created outside in order to be available for loading.

    Returns:
        str: The new ontology OWL file path within the SMIA Archive.
    """
    # The new ontology file is created in SMIA Archive inside configuration folder
    new_ontology_file_path = SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + "ontology.owl"
    with open(new_ontology_file_path, "wb") as binary_file:
        # Write bytes to file
        binary_file.write(file_bytes_content)
    return new_ontology_file_path

# --------------------------------------
# Methods related to asset information
# --------------------------------------
def get_asset_type():
    """
    This method returns the asset type of the AAS set in the ConfigMap by the AAS Controller during the deployment process. This information is stored in "asset.properties" file.

    Returns:
        str: The asset type of the AAS.
    """
    # Read submodels configuration
    config_sm = configparser.RawConfigParser()
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_ASSET_PROPERTIES_FILENAME)
    return config_sm['DEFAULT']['assetType']


# ----------------------------
# Methods related to submodels
# ----------------------------
def get_submodel_names():
    """
    This method returns all submodel names that have been selected for the AAS instance. It is the submodel
    properties file sections of the AAS configuration in the ConfigMap that will determine which submodels these
    are.

    Returns:
        list(str): A list with the sections of the submodel properties file, and therefore, the submodel names."""
    # Read submodels configuration
    config_sm = configparser.RawConfigParser()
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_SM_PROPERTIES_FILENAME)

    return config_sm.sections()


def get_submodel_information(submodel_name):
    """
    This method returns the submodel information of a specific submodel, from the submodel properties file of the
    configuration from the ConfigMap.

    Args:
        submodel_name  (str): The name of the submodel. To read from the submodel properties file, it is also
        the name of the section.

    Returns:
        dict: The submodel information in the same format as the submodel properties file content.
    """
    # Read submodels configuration
    config_sm = configparser.RawConfigParser()
    config_sm.read(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + SMIAGeneralInfo.CM_SM_PROPERTIES_FILENAME)

    return config_sm.items(submodel_name)



import basyx
from basyx.aas.model import Qualifier, HasSemantics, Namespace, Qualifiable

from smia.logic.exceptions import AASModelReadingError


class ExtendedQualifier(Qualifier):
    """This class contains methods to be added to Qualifier class of Basyx Python SDK model."""


class ExtendedHasSemantics(HasSemantics):
    """This class contains methods to be added to HasSemantics class of Basyx Python SDK model."""

    def check_semantic_id_exist(self, semantic_id_reference):
        """
        This method checks if a specific semanticID exists in an AAS meta-model element.

        Args:
            semantic_id_reference (str): semantic identifier.

        Returns:
            bool: result of the check (only True if the semanticID exists).
        """
        if self.semantic_id is None:
            return False
        for reference in self.semantic_id.key:
            if reference.value == semantic_id_reference:
                return True
        return False

    def check_suppl_semantic_id_exist(self, suppl_semantic_id_ref):
        """
        This method checks if a specific supplemental semanticID exists in an AAS meta-model element.

        Args:
            suppl_semantic_id_ref (str): supplemental semantic identifier.

        Returns:
            bool: result of the check (only True if the semanticID exists).
        """
        if self.supplemental_semantic_id is None:
            return False
        for suppl_semantic_id in self.supplemental_semantic_id:
            for reference in suppl_semantic_id.key:
                if reference.value == suppl_semantic_id_ref:
                    return True
            return False


class ExtendedNamespace(Namespace):

    def check_if_element_is_structural(self):
        """
        This method checks whether the AAS element is of type structural (Submodel, SubmodelElementCollection or
        SubmodelElementList).

        Returns:
            bool: result of the check
        """
        if isinstance(self, basyx.aas.model.Submodel) \
                or (isinstance(self, basyx.aas.model.SubmodelElementCollection)
                    or isinstance(self, basyx.aas.model.SubmodelElementList)):
            return True
        else:
            return False


class ExtendedQualifiable(Qualifiable):

    def get_qualifier_value_by_type(self, qualifier_type_value):
        """
        This method gets the value of the qualifier that has a given type.

        Args:
            qualifier_type_value (str): type of the qualifier.

        Returns:
            str: value of the qualifier with the given type
        """
        try:
            qualifier_object = self.get_qualifier_by_type(qualifier_type_value)
            if qualifier_object is None:
                raise AASModelReadingError("Qualifier type {} not found in the element {}".format(
                    qualifier_type_value, self), self, 'KeyError in qualifiers')
            else:
                return qualifier_object.value
        except KeyError as e:
            raise AASModelReadingError("Qualifier type {} not found in the element {}".format(
                qualifier_type_value, self), self, 'KeyError in qualifiers')

    def get_qualifier_value_by_semantic_id(self, qualifier_semantic_id):
        """
        This method gets the value of the qualifier that has a given semanticID.

        Args:
            qualifier_semantic_id (str): semanticID of the qualifier.

        Returns:
            str: value of the qualifier with the given type
        """
        try:
            if len(self.qualifier) == 0:
                raise AASModelReadingError("Qualifier not found in the element {} because the element has no "
                                           "qualifier with semanticID {}.".format(self, qualifier_semantic_id),
                                           self, 'KeyError in qualifiers')
            for qualifier in self.qualifier:
                if qualifier.check_semantic_id_exist(qualifier_semantic_id) is True:
                    return qualifier.value
            else:
                raise AASModelReadingError("Qualifier with semanticID {} not found in the element {}".format(
                    qualifier_semantic_id, self), self, 'SemanticIDError in qualifiers')

        except KeyError as e:
            raise AASModelReadingError("Qualifier with semanticID {} not found in the element {}".format(
                qualifier_semantic_id, self), self, 'KeyError in qualifiers')



import os


class SMIAGeneralInfo:
    """This class contains all general information about the SMIA."""

    # Variables related to states of the FSM
    # --------------------------------------
    BOOTING_STATE_NAME = 'BOOTING'
    RUNNING_STATE_NAME = 'RUNNING'
    STOPPING_STATE_NAME = 'STOPPING'
    IDLE_STATE_NAME = 'IDLE'

    # Path of the SMIA Archive
    SMIA_ARCHIVE_PATH = '/smia_archive'
    if 'KUBERNETES_PORT' not in os.environ: #  If the SMIA DT is run locally, the configuration files are in 'smia_archive' folder of the project
        SMIA_ARCHIVE_PATH = '../smia_archive'

    # Paths related to interactions between Manager and Core
    INTERACTIONS_FOLDER_PATH = '/aas_archive/interactions'
    CORE_INTERACTIONS_FOLDER_PATH = '/aas_archive/interactions/core'
    MANAGER_INTERACTIONS_FOLDER_PATH = '/aas_archive/interactions/manager'

    # Subpath of interaction files
    SVC_REQUEST_FILE_SUBPATH = '/svcRequests.json'
    SVC_RESPONSE_FILE_SUBPATH = '/svcResponses.json'

    # Paths for log files
    LOG_FOLDER_PATH = SMIA_ARCHIVE_PATH + '/log'
    SMIA_LOG_FILENAME = 'smia_dt.log'

    SVC_LOG_FOLDER_PATH = LOG_FOLDER_PATH + '/services'
    ERROR_LOG_FOLDER_PATH = LOG_FOLDER_PATH + '/errors'

    ASSET_RELATED_SVC_LOG_FILENAME = 'assetRelatedSvcHistory.log'
    AAS_INFRASTRUCTURE_SVC_LOG_FILENAME = 'aasInfrastructureSvcHistory.log'
    AAS_SERVICES_LOG_FILENAME = 'aasServicesHistory.log'
    SUBMODEL_SERVICES_LOG_FILENAME = 'submodelServicesHistory.log'
    SUBMODEL_CSS_LOG_FILENAME = 'capabilitySkillHistory.log'

    # ConfigMap files
    CONFIGURATION_FOLDER_PATH = SMIA_ARCHIVE_PATH + '/config'
    CONFIGURATION_AAS_FOLDER_PATH = CONFIGURATION_FOLDER_PATH + '/aas'
    CM_GENERAL_PROPERTIES_FILENAME = 'smia-initialization.properties'
    CM_AAS_MODEL_FILENAME = ''
    CM_AAS_PROPERTIES_FILENAME = 'aas.properties'
    CM_ASSET_PROPERTIES_FILENAME = 'asset.properties'
    CM_SM_PROPERTIES_FILENAME = 'submodels.properties'

    # Submodel files
    SUBMODEL_FOLDER_PATH = '/aas_archive/submodels'
    TECHNICAL_DATA_SM_FILENAME = 'Technical_data_SM.xml'
    CONFIGURATION_SM_FILENAME = 'Configuration_SM.xml'

    # Status file
    STATUS_FOLDER_PATH = SMIA_ARCHIVE_PATH + '/status'
    SMIA_STATUS_FILE_NAME = 'smia_dt.json'
    CORE_STATUS_FILE_PATH = '/aas_archive/status/aas_core.json'






import logging
from spade.behaviour import State

from smia.behaviours.negotiating_behaviour import NegotiatingBehaviour
from smia.behaviours.acl_handling_behaviour import ACLHandlingBehaviour
from smia.utilities import smia_archive_utils
from smia.utilities.general_utils import SMIAGeneralInfo
from smia.utilities.smia_info import SMIAInteractionInfo
from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyInfo

_logger = logging.getLogger(__name__)


class StateRunning(State):
    """
    This class contains the Running state of the common SMIA.
    """

    async def run(self):
        """
        This method implements the running state of the common SMIA. Here all requests services are handled,
        from ACL of another SMIA.
        """

        _logger.info("## STATE 2: RUNNING ##  (Initial state)")

        # SMIA is in the Running status
        smia_archive_utils.update_status('Running')

        # On the one hand, a behaviour is required to handle ACL messages
        acl_handling_behav = ACLHandlingBehaviour(self.agent)
        self.agent.add_behaviour(acl_handling_behav, SMIAInteractionInfo.SVC_STANDARD_ACL_TEMPLATE)

        # On the other hand, a behaviour is required to handle interaction messages
        # TODO revisar, ya que en el nuevo enfoque no hay AAS Core
        # interaction_handling_behav = InteractionHandlingBehaviour(self.agent)
        # self.agent.add_behaviour(interaction_handling_behav)

        # Besides, the negotiation behaviour has to be added to the agent
        agent_behaviours_classes = await self.add_agent_capabilities_behaviours()

        # Wait until the behaviour has finished. Is a CyclicBehaviour, so it will not end until an error occurs or, if
        # desired, it can be terminated manually using "behaviour.kill()".
        await acl_handling_behav.join()
        # await interaction_handling_behav.join()
        if agent_behaviours_classes:
            # TODO revisar si esto se quiere hacer asi (pensar en las transciones entre estados)
            for behav_class in agent_behaviours_classes:
                await behav_class.join()

        # If the Execution Running State has been completed, the agent can move to the next state
        _logger.info(f"{self.agent.jid} agent has finished it Running state.")
        self.set_next_state(SMIAGeneralInfo.STOPPING_STATE_NAME)

    async def add_agent_capabilities_behaviours(self):
        """
        This method adds all the behaviors associated to the agent capabilities. In case of being an ExtensibleAgent,
        it is necessary to analyze if new behaviors have been added through the extension mechanisms.

        Returns:
            behaviours_instances: all instances of behavior to know that these are part of the Running state.
        """
        behaviours_instances = []
        agent_capabilities = await self.agent.css_ontology.get_ontology_instances_by_class_iri(
            CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI)
        for capability_instance in agent_capabilities:
            if capability_instance.name == 'Negotiation':
                # The negotiation behaviour has to be added to the agent
                _logger.info("This SMIA has negotiation capability.")
                negotiation_behav = NegotiatingBehaviour(self.agent)
                self.agent.add_behaviour(negotiation_behav, SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE)
                behaviours_instances.append(negotiation_behav)
            elif capability_instance.name == 'OtherAgentCapability':
                # TODO pensarlo
                pass

        from smia.agents.extensible_smia_agent import ExtensibleSMIAAgent  # Local import to avoid circular import error
        if isinstance(self.agent, ExtensibleSMIAAgent):
            if len(self.agent.extended_agent_capabilities) != 0:
                _logger.info("Extended agent capabilities will be added for the ExtensibleSMIAAgent.")
                for behav_instance in self.agent.extended_agent_capabilities:
                    self.agent.add_behaviour(behav_instance)
                    behaviours_instances.append(behav_instance)

        return behaviours_instances
